/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.admin;

import com.funambol.framework.tools.beans.BeanException;
import com.funambol.framework.tools.beans.BeanInstantiationException;
import com.funambol.framework.tools.beans.BeanInitializationException;
import com.funambol.framework.tools.beans.BeanNotFoundException;

import com.funambol.server.config.Configuration;

import com.funambol.exchange.exception.ExchangeConfigException;

/**
 * This class loads the parameter from the XML configuration file
 *
 * @version $Id: ExchangeConnectorConfig.java,v 1.1 2006/08/08 07:52:29 gibi Exp $
 */
public class ExchangeConnectorConfig {

    // --------------------------------------------------------------- Constants

    public final static String beanName = "exc/exc/FunambolExchangeConnector.xml";

    // -------------------------------------------------------------- Properties

    /** secc server         */
    private String    seccServer ;

    /** secc port           */
    private int       seccPort;

    /** exchange serve      */
    private String    exchangeServer;

    /** SSL activation */
    private boolean useSSL;

    /** keystore file name */
    private String keyStoreFileName;

    /** keystore password */
    private String keyStorePassword;

    /** filter activation    */
    //private boolean   filter;

    //------------------------------------------------------------- Constructors
    public ExchangeConnectorConfig () {
    }

    //----------------------------------------------------------- Public Methods

   /**
    *
    * @return ExchangeConnectorConfig
    * @throws ExchangeConfigException
    */
   public static ExchangeConnectorConfig getConfigInstance()
   throws ExchangeConfigException {

       ExchangeConnectorConfig exchangeConf;
       try {
          exchangeConf = (ExchangeConnectorConfig)Configuration.getConfiguration().getBeanInstanceByName(beanName);
       } catch (BeanInstantiationException bie){
          throw new ExchangeConfigException(bie.getMessage());
       } catch (BeanInitializationException bie){
          throw new ExchangeConfigException(bie.getMessage());
       } catch (BeanNotFoundException bie){
          throw new ExchangeConfigException(bie.getMessage());
       } catch (BeanException be){
          throw new ExchangeConfigException(be.getMessage());
       }

       return exchangeConf;

    }

    /**
     *
     *
     * @param seccServer String
     */
    public void setSeccServer(String seccServer){
        this.seccServer = seccServer;
    }

    /**
     *
     *
     * @return seccServer String
     */
    public String getSeccServer(){
        return this.seccServer;
    }


    /**
     *
     *
     * @param seccPort int
     */
    public void setSeccPort(int seccPort){
        this.seccPort = seccPort;
    }

    /**
     *
     *
     * @return seccPort int
     */
    public int getSeccPort(){
        return this.seccPort;
    }

    /**
     *
     *
     * @param exchangeServer String
     */
    public void setExchangeServer(String exchangeServer){
        this.exchangeServer = exchangeServer;
    }

    /**
     *
     *
     * @return exchangeServer
     */
    public String getExchangeServer(){
        return this.exchangeServer;
    }

    /**
     * <p>Getter of the useSSL flag of the Exchange Connector Configuration</p>
     * @return the value of the useSLL property
     */
    public boolean getUseSSL() {
        return useSSL;
    }

    /**
     * <p>Setter of the useSSL flag of the Exchange Connector Configuration</p>
     * @param useSSL
     */
    public void setUseSSL(boolean useSSL) {
        this.useSSL = useSSL;
    }

    /**
     * <p>Getter for the keystore's file name</p>
     * @return value of the property keyStoreFileName
     */
    public String getKeyStoreFileName() {
        return keyStoreFileName;
    }

    /**
     * <p>Setter of the property keyStoreFileName</p>
     * @param keyStoreFileName
     */
    public void setKeyStoreFileName(String keyStoreFileName) {
        this.keyStoreFileName = keyStoreFileName;
    }

    /**
     * <p>Getter of the property keyStorePassword</p>
     * @return value of the property keyStorePassword
     */
    public String getKeyStorePassword() {
        return keyStorePassword;
    }

    /**
     * Setter of the property keyStorePassword
     * @param keyStorePassword
     */
    public void setKeyStorePassword(String keyStorePassword) {
        this.keyStorePassword = keyStorePassword;
    }

}