/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.admin;

import java.io.Serializable;

import java.util.StringTokenizer;

import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.border.TitledBorder;

import org.apache.commons.lang.StringUtils;

import com.funambol.framework.engine.source.ContentType;
import com.funambol.framework.engine.source.SyncSourceInfo;

import com.funambol.admin.AdminException;
import com.funambol.admin.mo.SyncSourceManagementObject;
import com.funambol.admin.ui.SourceManagementPanel;

import com.funambol.exchange.engine.source.ExchangeCalendarSyncSource;
import com.funambol.exchange.engine.source.ExchangeContactSyncSource;
import com.funambol.exchange.engine.source.ExchangeNoteSyncSource;
import com.funambol.exchange.engine.source.ExchangeTaskSyncSource;
import com.funambol.exchange.engine.source.ExchangeSyncSource;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import javax.swing.JCheckBox;

/**
 * This class implements the configuration panel for an ExchangeSyncSource
 *
 * @version $Id: ExchangeSyncSourceConfigPanel.java,v 1.1 2006/08/08 07:52:29 gibi Exp $
 */
public class ExchangeSyncSourceConfigPanel
extends SourceManagementPanel
implements Serializable {

  // ----------------------------------------------------------------- Constants

  public static final String NAME_ALLOWED_CHARS =
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890-_."     ;
  public static final String PORT_ALLOWED_CHARS =
    "0123456789"                                                            ;

  public static final String VCARD_TYPE         = "text/x-vcard"            ;
  public static final String VCARD_TYPES        = "text/x-vcard"            ;
  public static final String VCARD_VERSIONS     = "2.1"                     ;
  public static final String ICAL_TYPE          = "text/x-vcalendar"        ;
  public static final String ICAL_TYPES         = "text/x-vcalendar"        ;
  public static final String ICAL_VERSIONS      = "1.0"                     ;
  public static final String SIFC_TYPE          = "text/x-s4j-sifc"         ;
  public static final String SIFC_TYPES         = "text/x-s4j-sifc"         ;
  public static final String SIFC_VERSIONS      = "1.0"                     ;
  public static final String SIFE_TYPE          = "text/x-s4j-sife"         ;
  public static final String SIFE_TYPES         = "text/x-s4j-sife"         ;
  public static final String SIFE_VERSIONS      = "1.0";
  public static final String SIFN_TYPE          = "text/x-s4j-sifn"         ;
  public static final String SIFN_TYPES         = "text/x-s4j-sifn"         ;
  public static final String SIFN_VERSIONS      = "1.0";
  public static final String SIFT_TYPE          = "text/x-s4j-sift"         ;
  public static final String SIFT_TYPES         = "text/x-s4j-sift"         ;
  public static final String SIFT_VERSIONS      = "1.0"                     ;

  // -------------------------------------------------------------- Private data
  /** label for the panel's name */
  private JLabel             panelName            = new JLabel        () ;
  private JLabel             nameLabel            = new JLabel        () ;
  private JTextField         nameValue            = new JTextField    () ;
  private JLabel             sourceUriLabel       = new JLabel        () ;
  private JTextField         sourceUriValue       = new JTextField    () ;
  private JLabel             exchangeFolderLabel  = new JLabel        () ;
  private JTextField         exchangeFolderValue  = new JTextField    () ;
  private JLabel             conversionTypeLabel  = new JLabel        () ;
  private JComboBox          conversionValue      = new JComboBox     () ; 
  private JLabel             dtEncryptLabel       = new JLabel        () ;
  private JCheckBox          dtEncryptValue       = new JCheckBox     () ;
  private JLabel             dtEncodeLabel        = new JLabel        () ;
  private JCheckBox          dtEncodeValue        = new JCheckBox     () ;  
  private JButton            confirmButton        = new JButton       () ;

  // -------------------------------------------------------------- Constructors

  public ExchangeSyncSourceConfigPanel() {
      init();
  }

  // ----------------------------------------------------------- Private methods
  /**
   * Create the panel
   */

  private void init()
  {
    TitledBorder titledBorder1;
    
    // set layout
    this.setLayout(null);

    // set properties of label, position and border
    //  referred to the title of the panel
    titledBorder1 = new TitledBorder("");

    panelName.setFont(titlePanelFont);
    panelName.setText("Edit Exchange SyncSource");
    panelName.setBounds(new Rectangle(14, 5, 316, 28));
    panelName.setAlignmentX(SwingConstants.CENTER);
    panelName.setBorder(titledBorder1);

    sourceUriLabel.setText("Source URI: ");
    sourceUriLabel.setFont(defaultFont);
    sourceUriLabel.setBounds(new Rectangle(14, 60, 150, 18));
    sourceUriValue.setFont(defaultFont);
    sourceUriValue.setBounds(new Rectangle(170, 60, 350, 18));

    nameLabel.setText("Name: ");
    nameLabel.setFont(defaultFont);
    nameLabel.setBounds(new Rectangle(14, 90, 150, 18));
    nameValue.setFont(defaultFont);
    nameValue.setBounds(new Rectangle(170, 90, 350, 18));

    exchangeFolderLabel.setText("Exchange Folder: ");
    exchangeFolderLabel.setFont(defaultFont);
    exchangeFolderLabel.setBounds(new Rectangle(14, 120, 150, 18));
    exchangeFolderValue.setFont(defaultFont);
    exchangeFolderValue.setBounds(new Rectangle(170, 120, 350, 18));

    conversionTypeLabel.setText("Type: ");
    conversionTypeLabel.setFont(defaultFont);
    conversionTypeLabel.setBounds(new Rectangle(14, 150, 150, 18));
    conversionValue.setFont(defaultFont);
    conversionValue.setBounds(new Rectangle(170, 150, 80, 18));

    conversionValue.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent ie) {
            if (conversionValue.getSelectedIndex() == 0){
               dtEncryptValue.setSelected(false);
               dtEncodeValue.setSelected(true);
            } else {
               dtEncryptValue.setSelected(false);
               dtEncodeValue.setSelected(false);
            }   
        }
    });
    
    
    
    dtEncryptLabel.setText("encryption: ");
    dtEncryptLabel.setFont(defaultFont);
    dtEncryptLabel.setBounds(new Rectangle(170, 180, 70, 18));
    dtEncryptValue.setSelected(false);
    dtEncryptValue.setBounds(new Rectangle(250, 180, 25, 19));
  
    dtEncodeLabel.setText("encoding: ");
    dtEncodeLabel.setFont(defaultFont);
    dtEncodeLabel.setBounds(new Rectangle(280, 180, 70, 18));
    dtEncodeValue.setSelected(false);
    dtEncodeValue.setBounds(new Rectangle(360, 180, 25, 19));
    dtEncodeValue.setEnabled(false);
    
    dtEncryptValue.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
            if (e.getSource().equals(dtEncryptValue)) {
                if (dtEncryptValue.isSelected()){
                   dtEncodeValue.setSelected(true);                   
                } else {
                    if (conversionValue.getSelectedIndex() == 1){
                       dtEncodeValue.setSelected(false);                   
                    } else {
                       dtEncodeValue.setSelected(true);                   
                    }
                }   
            }
        }
    });
        
    confirmButton.setFont(defaultFont);
    confirmButton.setText("Add");
    confirmButton.setBounds(170, 220, 70, 25);

    confirmButton.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent event ) {
            try {
                validateValues();
                getValues();                
                if (getState() == STATE_INSERT) {
                    ExchangeSyncSourceConfigPanel.this.actionPerformed(
                            new ActionEvent(ExchangeSyncSourceConfigPanel.this, 
                                            ACTION_EVENT_INSERT, 
                                            event.getActionCommand()));
                } else {
                    ExchangeSyncSourceConfigPanel.this.actionPerformed(
                            new ActionEvent(ExchangeSyncSourceConfigPanel.this, 
                                            ACTION_EVENT_UPDATE, 
                                            event.getActionCommand()));
                }
            } catch (Exception e) {
                notifyError(new AdminException(e.getMessage()));
            }
        }
    });

    // add all components to the panel
    this.add(panelName            , null ) ;
    this.add(sourceUriLabel       , null ) ;
    this.add(sourceUriValue       , null ) ;
    this.add(nameLabel            , null ) ;
    this.add(nameValue            , null ) ;
    this.add(exchangeFolderLabel  , null ) ;
    this.add(exchangeFolderValue  , null ) ;
    this.add(conversionTypeLabel  , null ) ;
    this.add(conversionValue      , null ) ;
    this.add(dtEncryptLabel       , null ) ;
    this.add(dtEncryptValue       , null ) ;
    this.add(dtEncodeLabel        , null ) ;
    this.add(dtEncodeValue        , null ) ;
    this.add(confirmButton        , null ) ;

  }

  /**
   *
   *
   */
  public void updateForm() {
      
      if (!(getSyncSource() instanceof ExchangeSyncSource)) {
          notifyError(new AdminException(
              "This is not an ExchangeSyncSource! Unable to process SyncSource values."));
          return;
      }
      
      ExchangeSyncSource syncSource = (ExchangeSyncSource)getSyncSource();

      if (getState() == STATE_INSERT) {
          confirmButton.setText("Add");
      } else if (getState() == STATE_UPDATE) {
          confirmButton.setText("Save");
      }

      sourceUriValue.setText(syncSource.getSourceURI());
      exchangeFolderValue.setText(syncSource.getExchangeFolder());
      nameValue.setText(syncSource.getName());

      conversionValue.removeAllItems();

      if (syncSource instanceof ExchangeCalendarSyncSource)  {
        conversionValue.addItem ("SIF-E" );
        conversionValue.addItem ("iCal"  );
      } else if (syncSource instanceof ExchangeContactSyncSource) {
        conversionValue.addItem ("SIF-C" );
        conversionValue.addItem ("vCard" );
      } else if (syncSource instanceof ExchangeNoteSyncSource) {
        conversionValue.addItem ("SIF-N" );
      }else if (syncSource instanceof ExchangeTaskSyncSource) {
        conversionValue.addItem ("SIF-T" );
      }

      if ((syncSource instanceof ExchangeCalendarSyncSource) ||
          (syncSource instanceof ExchangeContactSyncSource)) {
          String cType = syncSource.getType();
          if (VCARD_TYPE.equals(cType) || ICAL_TYPE.equals(cType)) {
              conversionValue.setSelectedIndex(1);
          } else {
              conversionValue.setSelectedIndex(0);
          }
      }

      SyncSourceManagementObject mo = (SyncSourceManagementObject)getManagementObject();
      String transformationsRequired = mo.getTransformationsRequired();
      //notifyError(new AdminException("TransformationsRequired: " + transformationsRequired));
      if (transformationsRequired == null){
          dtEncryptValue.setSelected(false);
          dtEncodeValue.setSelected(false);
      } else {
          if(transformationsRequired.equals("des;b64")){
              dtEncryptValue.setSelected(true);
              dtEncodeValue.setSelected(true);
          } else if (transformationsRequired.equals("b64")){
              dtEncryptValue.setSelected(false);
              dtEncodeValue.setSelected(true);
          } else {
              dtEncryptValue.setSelected(false);
              dtEncodeValue.setSelected(false);
          }
      }
      if (conversionValue.getSelectedIndex() == 0){
         dtEncodeValue.setSelected(true);
      }       
            
      if (syncSource.getSourceURI() != null) {
          sourceUriValue.setEditable(false);
      }
      
  }

  // ----------------------------------------------------------- Private methods

  /**
   * Checks if the values provided by the user are all valid. In caso of errors,
   * a IllegalArgumentException is thrown.
   *
   * @throws IllegalArgumentException if:
   *         <ul>
   *         <li>name, uri, type or directory are empty (null or zero-length)
   *         <li>the types list length does not match the versions list length
   *         </ul>
   */
  private void validateValues() throws IllegalArgumentException {
      String value;

      value = nameValue.getText();
      if (StringUtils.isEmpty(value)) {
          throw new
            IllegalArgumentException(
                "Field 'Name' cannot be empty. "    +
                "Please provide a SyncSource name." );
      }

      if (!StringUtils.containsOnly(value, NAME_ALLOWED_CHARS.toCharArray())) {
          throw new
            IllegalArgumentException(
                "Only the following characters are "   +
                "allowed for field 'Name': \n"         +
                NAME_ALLOWED_CHARS                     );
      }

      value = exchangeFolderValue.getText();
      if (StringUtils.isEmpty(value)) {
        throw new
            IllegalArgumentException(
                "Field 'Exchange folder' cannot be empty." +
                "Please provide an Exchange folder."       );
      }

      value = sourceUriValue.getText();
      if (StringUtils.isEmpty(value)) {
          throw new
            IllegalArgumentException(
                "Field 'Source URI' cannot be empty. "  +
                "Please provide a SyncSource URI."      );
      }

  }

  /**
   * Set syncSource properties with the values provided by the user.
   */
  private void getValues() {

        ExchangeSyncSource syncSource = (ExchangeSyncSource)getSyncSource();

        StringTokenizer types    = null ;
        StringTokenizer versions = null ;

        syncSource.setSourceURI(sourceUriValue.getText().trim()) ;
        syncSource.setExchangeFolder(exchangeFolderValue.getText().trim()) ;
        syncSource.setName(nameValue.getText().trim()) ;
                
        if (conversionValue.getSelectedIndex() == 0) {
            // SIF
            if (syncSource instanceof ExchangeCalendarSyncSource)        {
                syncSource.setType(SIFE_TYPE);
                types = new StringTokenizer(SIFE_TYPES       , "," ) ;
                versions = new StringTokenizer(SIFE_VERSIONS , "," ) ;
            } else if (syncSource instanceof ExchangeContactSyncSource ) {
                syncSource.setType(SIFC_TYPE);
                types = new StringTokenizer(SIFC_TYPES       , "," ) ;
                versions = new StringTokenizer(SIFC_VERSIONS , "," ) ;
            } else if (syncSource instanceof ExchangeNoteSyncSource )    {
                syncSource.setType(SIFN_TYPE);
                types = new StringTokenizer(SIFN_TYPES       , "," ) ;
                versions = new StringTokenizer(SIFN_VERSIONS , "," ) ;
            } else if (syncSource instanceof ExchangeTaskSyncSource )    {
                syncSource.setType(SIFT_TYPE);
                types = new StringTokenizer(SIFT_TYPES       , "," ) ;
                versions = new StringTokenizer(SIFT_VERSIONS , "," ) ;
            }
        } else if (syncSource instanceof ExchangeCalendarSyncSource)     {
            // ICAL
            syncSource.setType(ICAL_TYPE);
            types = new StringTokenizer(ICAL_TYPES       , "," ) ;
            versions = new StringTokenizer(ICAL_VERSIONS , "," ) ;
        } else if (syncSource instanceof ExchangeContactSyncSource )     {
            // VCARD
            syncSource.setType(VCARD_TYPE);
            types = new StringTokenizer(VCARD_TYPES       , "," ) ;
            versions = new StringTokenizer(VCARD_VERSIONS , "," ) ;
        }

       
        boolean dtencr = dtEncryptValue.isSelected();
        boolean dtenco = dtEncodeValue.isSelected();        
        String transformationsRequired = null;    
        if(dtencr){
           transformationsRequired = "des;b64"; 
        } else if (!dtencr && dtenco){
           transformationsRequired = "b64"; 
        } else {
           transformationsRequired = null; 
        }
        SyncSourceManagementObject mo = (SyncSourceManagementObject)getManagementObject();
        mo.setTransformationsRequired(transformationsRequired);                                

        
        ContentType[] contentTypes= new ContentType[types.countTokens()];

        for(int i = 0, l = contentTypes.length; i < l; ++i) {
          contentTypes[i] = new ContentType(types.nextToken().trim()    ,
                                            versions.nextToken().trim() );
        }

        syncSource.setInfo(new SyncSourceInfo(contentTypes, 0));
  }

}