/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.items.calendar.dao;

import java.text.MessageFormat;
import java.text.SimpleDateFormat;
import java.text.DateFormat;
import java.text.ParseException;

import java.util.*;

import com.funambol.foundation.pdi.event.Reminder;
import com.funambol.foundation.pdi.event.RecurrencePattern;

import com.funambol.exchange.DataAccessException;
import com.funambol.exchange.httptransport.WebDavHttpTransport;
import com.funambol.exchange.items.calendar.model.Calendar;
import com.funambol.exchange.items.common.dao.ItemDAO;
import com.funambol.exchange.util.DateTools;
import com.funambol.exchange.util.StringTools;
import com.funambol.exchange.xml.XmlParser;
import com.funambol.exchange.xml.XmlParseException;

/*
 * This class implements methods to access calendar data
 * in exchange server datastore
 *
 * @version $Id: CalendarDAO.java,v 1.2 2006/09/14 08:54:52 adeac Exp $
 *
 **/

public class CalendarDAO extends ItemDAO {

    //---------------------------------------------------------------- Constants
    public static final String TAG_INSTANCETYPE = "instancetype";
    public static final String TAG_ALLDAYEVENT = "alldayevent";
    public static final String TAG_BODY = "textdescription";
    public static final String TAG_BUSY_STATUS = "busystatus";
    public static final String TAG_DATEEND = "dtend";
    public static final String TAG_DATESTART = "dtstart";

    public static final String TAG_LOCATION = "location";
    public static final String TAG_RESPONSE = "a:response";
    public static final String TAG_SUBJECT = "subject";

    public static final String TAG_REMINDER_SET = "reminderset";
    public static final String TAG_REMINDER_OFFSET = "reminderoffset";

    public static final String TAG_REPLY_TIME = "replytime";

    public static final String TAG_SENSITIVITY = "sensitivity";

    public static final String TAG_RECURRENCE = "rrule";

    public static final String TAG_SEQUENCE = "sequence";

    public static final String TAG_MILEAGE = "mileage";
    public static final String TAG_IMPORTANCE = "importance";

    public static final String TAG_CATEGORIES = "categories";

    public static final String TAG_REPLUID = "repluid";
    public static final String TAG_LAST_MODIFIED = "getlastmodified";
    public static final String TAG_APPT_STATE_FLAGS = "apptstateflags";

    public static final String TAG_TIMEZONE_ID = "timezoneid";

    private static final String WEBDAV_MSG_SELECT_CALENDARS =
            "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n" +
                    "<D:searchrequest xmlns:D =\"DAV:\">\n" +
                    "<D:sql>\n" +
                    "Select " +
                    "\"http://schemas.microsoft.com/repl/repl-uid\" AS " + TAG_REPLUID + ", " +
                    "\"urn:schemas:calendar:location\" AS " + TAG_LOCATION + ", " +
                    "\"urn:schemas:httpmail:subject\" AS " + TAG_SUBJECT + ", " +
                    "\"urn:schemas:httpmail:textdescription\" AS " + TAG_BODY + ", " +
                    "\"urn:schemas:calendar:dtstart\" AS " + TAG_DATESTART + ", " +
                    "\"urn:schemas:calendar:dtend\" AS " + TAG_DATEEND + ", " +
                    "\"urn:schemas:calendar:alldayevent\" AS " + TAG_ALLDAYEVENT + ", " +
                    "\"urn:schemas:calendar:busystatus\" AS " + TAG_BUSY_STATUS + ", " +
                    "\"urn:schemas:calendar:sequence\" AS " + TAG_SEQUENCE + ", " +
                    "\"urn:schemas:calendar:rrule\" AS " + TAG_RECURRENCE + ", " +
                    "\"http://schemas.microsoft.com/mapi/reminderset\" AS " + TAG_REMINDER_SET + ", " +
                    "\"urn:schemas:calendar:reminderoffset\" AS " + TAG_REMINDER_OFFSET + ", " +
                    "\"urn:schemas:calendar:replytime\" AS " + TAG_REPLY_TIME + ", " +
                    "\"http://schemas.microsoft.com/mapi/sensitivity\" AS " + TAG_SENSITIVITY + ", " +
                    "\"http://schemas.microsoft.com/exchange/mileage\" AS " + TAG_MILEAGE + ", " +
                    "\"http://schemas.microsoft.com/exchange/keywords-utf8\" AS " + TAG_CATEGORIES + ", " +
                    "\"urn:schemas:httpmail:importance\" AS " + TAG_IMPORTANCE + ", " +
                    "\"DAV:getlastmodified\" AS " + TAG_LAST_MODIFIED + ", " +
                    "\"DAV:href\" AS href, " +
                    "\"DAV:isfolder\" AS isfolder, " +
                    "\"urn:schemas:calendar:instancetype\" AS instancetype " +
                    "FROM \"/{0}/{1}/{2}\"  " +
                    "{3}" +
                    "</D:sql>\n" +
                    "</D:searchrequest>";

    private static final String WEBDAV_MSG_SELECT_CALENDARS_FOR_TWIN =
            "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n" +
                    "<D:searchrequest xmlns:D =\"DAV:\">\n" +
                    "<D:sql>\n" +
                    "Select " +
                    "\"http://schemas.microsoft.com/repl/repl-uid\" AS repluid, " +
                    "\"DAV:getlastmodified\" AS getlastmodified, " +
                    "\"DAV:creationdate\" AS creationdate, " +
                    "\"DAV:href\" AS href, " +
                    "\"urn:schemas:calendar:alldayevent\" AS "
                    + TAG_ALLDAYEVENT + ", " +
                    "\"urn:schemas:calendar:location\" AS "
                    + TAG_LOCATION + ", " +
                    "\"urn:schemas:calendar:dtstart\" AS "
                    + TAG_DATESTART + ", " +
                    "\"urn:schemas:calendar:dtend\" AS "
                    + TAG_DATEEND + ", " +
                    "\"DAV:isfolder\" AS isfolder " +
                    "FROM \"/{0}/{1}/{2}\"  " +
                    "{3}" +
                    "</D:sql>\n" +
                    "</D:searchrequest>";

    private static final String FIELD_PREFIX = "urn:schemas:calendar:";

    private static final String
            WEBDAV_INIT_HM = "HM:";
    private static final String
            WEBDAV_INIT_HN = "HN:";
    private static final String
            WEBDAV_INIT_MAPI = "MAPI:";
    private static final String
            WEBDAV_INIT_EXCHANGE = "EX:";

    private static final String
            WEBDAV_TAG_INSTANCETYPE = WEBDAV_INIT_HM +
            TAG_INSTANCETYPE;

    private static final String
            WEBDAV_TAG_ALLDAY = WEBDAV_INIT_HM +
            TAG_ALLDAYEVENT;
    private static final String
            WEBDAV_TAG_BODY = WEBDAV_INIT_HN +
            TAG_BODY;
    private static final String
            WEBDAV_TAG_BUSYSTATUS = WEBDAV_INIT_HM +
            TAG_BUSY_STATUS;
    private static final String
            WEBDAV_TAG_DATEEND = WEBDAV_INIT_HM +
            TAG_DATEEND;
    private static final String
            WEBDAV_TAG_DATESTART = WEBDAV_INIT_HM +
            TAG_DATESTART;
    private static final String
            WEBDAV_TAG_LOCATION = WEBDAV_INIT_HM +
            TAG_LOCATION;
    private static final String
            WEBDAV_TAG_SUBJECT = WEBDAV_INIT_HN +
            TAG_SUBJECT;
    private static final String
            WEBDAV_TAG_IMPORTANCE = WEBDAV_INIT_HN +
            TAG_IMPORTANCE;
    private static final String
            WEBDAV_TAG_REMINDER_SET = WEBDAV_INIT_MAPI +
            TAG_REMINDER_SET;
    private static final String
            WEBDAV_TAG_REMINDER_OFFSET = WEBDAV_INIT_HM +
            TAG_REMINDER_OFFSET;
    private static final String
            WEBDAV_TAG_REPLY_TIME = WEBDAV_INIT_HM +
            TAG_REPLY_TIME;
    private static final String
            WEBDAV_TAG_SENSITIVITY = WEBDAV_INIT_MAPI +
            TAG_SENSITIVITY;
    private static final String WEBDAV_TAG_APPT_STATE_FLAGS =
            WEBDAV_INIT_MAPI + TAG_APPT_STATE_FLAGS;
    private static final String
            WEBDAV_TAG_RECURRENCE = WEBDAV_INIT_HM +
            TAG_RECURRENCE;
    private static final String
            WEBDAV_TAG_SEQUENCE = WEBDAV_INIT_HM +
            TAG_SEQUENCE;
    private static final String
            WEBDAV_TAG_MILEAGE = WEBDAV_INIT_EXCHANGE +
            TAG_MILEAGE;
    private static final String
            WEBDAV_TAG_TIMEZONE_ID = WEBDAV_INIT_HM + TAG_TIMEZONE_ID;
    private static final String
            EVERY_WEEKDAY = "MO, TU, WE, TH, FR";

    //------------------------------------------------------------- Private data


    private WebDavHttpTransport webDavHttp = null;

    // The timezone to use (if required) in the dates conversion
    private TimeZone timezone = null;

    //-------------------------------------------------------------- Costructors

    /**
     * <p>Constructor of the CalendarDAO class.</p>
     *
     * @param host           the host to which the Connector directlly connects (i. e. SECC Gate)
     * @param port           the port to which the Connector connects
     * @param exchangeServer the address of the Exchange Server
     * @param timezone       the timezone of the used device
     * @throws DataAccessException if an error occurs
     */
    public CalendarDAO(String host,
                       int port,
                       String exchangeServer,
                       TimeZone timezone)
            throws DataAccessException {

        this.webDavHttp = new WebDavHttpTransport(host,
                port,
                exchangeServer);

        this.timezone = timezone;
    }
    //----------------------------------------------------------- Public methods

    /**
     * insert / update calendar<br>
     *
     * @param c
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @return new / updated calendar
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    public Calendar setCalendar(Calendar c,
                                String username,
                                String credentials,
                                String exchangeFolder)
            throws DataAccessException {

        String response;

        String id;

        StringBuffer webDavCalendarMsg;
        String webDavHeaderMsg;

        String startDate;
        String endDate;
        String subject = null;
        String body = null;
        String location = null;
        Boolean allDayB;
        String allDay = null;
        String busyStatus;

        String importance = null;
        Integer mileage;

        int reminderMinutesBeforeStart = 0;
        boolean reminderSet = false;
        String replyTime = null;
        String sensitivity = null;
        String sequence = null;

        String categories = null;

        String server;
        String resource;

        String recurrence = null;

        int timezoneId;

        //String       rrule;

        RecurrencePattern rp = c.getRecurrencePattern();

        server = getServerFromExchangeFolder(exchangeFolder);
        resource = getResourceFromExchangeFolder(exchangeFolder);

        //rrule = c.getRrule().getPropertyValueAsString();

        try {

            //
            // We have to check the date format because if the event is an
            // allDayEvent, the format is yyyy-MM-dd
            //
            boolean isAllDayEvent = c.getAllDay().booleanValue();

            if (!isAllDayEvent) {
                startDate = DateTools.dateToWebDavTag(c.getDtStart().getPropertyValueAsString(),
                        TimeZone.getTimeZone("GMT"));
                endDate = DateTools.dateToWebDavTag(c.getDtEnd().getPropertyValueAsString(),
                        TimeZone.getTimeZone("GMT"));
            } else {
                //
                // These are yyyy-MM-dd
                //
                startDate = c.getDtStart().getPropertyValueAsString();
                endDate = c.getDtEnd().getPropertyValueAsString();

                startDate = DateTools.convertDayInWebDavDateFormat(startDate, timezone, true);
                endDate = DateTools.convertDayInWebDavDateFormat(endDate, timezone, false);

            }

            if (rp != null) {
                try {
                    recurrence = getExchangeRecurrence(rp, startDate);
                } catch (Exception e) {
                    throw new DataAccessException(e.getMessage(), e);
                }

            }

        } catch (Exception e) {
            throw new DataAccessException(e.getMessage(), e);
        }

        if (c.getSummary() != null) {
            subject = c.getSummary().getPropertyValueAsString();
        }

        if (c.getDescription() != null) {
            body = c.getDescription().getPropertyValueAsString();
        }

        if (c.getLocation() != null) {
            location = c.getLocation().getPropertyValueAsString();
        }

        if (c.getCategories() != null) {
            categories = c.getCategories().getPropertyValueAsString();
        }

        allDayB = c.getAllDay();

        if (allDayB != null) {

            if (allDayB.booleanValue()) {
                allDay = "1";
            } else {
                allDay = "0";
            }
        }

        if (c.getPriority() != null) {
            importance = c.getPriority().getPropertyValueAsString();
        }
        if (importance == null) {
            importance = "1";
        }

        mileage = c.getMileage();

        Reminder reminder = c.getReminder();
        if (reminder != null) {
            reminderMinutesBeforeStart = reminder.getMinutes();
            reminderSet = reminder.isActive();
        }

        try {

            if (c.getReplyTime() != null &&
                    c.getReplyTime().getPropertyValueAsString() != null &&
                    c.getReplyTime().getPropertyValueAsString().length() > 0) {

                replyTime = DateTools.dateToWebDavTag(c.getReplyTime().getPropertyValueAsString(),
                        TimeZone.getTimeZone("GMT"));

            }

        } catch (Exception e) {
            throw new DataAccessException(e.getMessage());
        }

        if (c.getClassEvent() != null) {
            sensitivity = c.getClassEvent().getPropertyValueAsString();
        }

        if (c.getSequence() != null) {
            sequence = c.getSequence().getPropertyValueAsString();
        }

        short sBusyStatus = -1;
        if (c.getBusyStatus() != null) {
            sBusyStatus = c.getBusyStatus().shortValue();
        }

        switch (sBusyStatus) {
            case 0: {
                busyStatus = "Free";
                break;
            }
            case 1: {
                busyStatus = "Tentative";
                break;
            }
            case 2: {
                busyStatus = "Busy";
                break;
            }
            case 3: {
                busyStatus = "OOF";
                break;
            }
            default: {
                busyStatus = null;
            }
        }

        timezoneId = getTimezoneId(timezone);

        webDavCalendarMsg = new StringBuffer(INITIAL_SIZE_SBUFFER);

        webDavCalendarMsg.
                append("<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n")
                .append("<D:propertyupdate xmlns:D=\"DAV:\" ")
                .append("xmlns:EX=\"http://schemas.microsoft.com/exchange/\" ")
                .append("xmlns:HM=\"urn:schemas:calendar:\" ")
                .append("xmlns:MAPI=\"http://schemas.microsoft.com/mapi/\" ")
                .append("xmlns:X=\"xml:\" ")
                .append("xmlns:HN=\"urn:schemas:httpmail:\" ")
                .append("xmlns:dt=\"urn:uuid:c2f41010-65b3-11d1a29f-00aa00c14882/\" >\n")
                .append("<D:set>\n")
                .append("<D:prop>\n")
                .append("<D:contentclass>urn:content-classes:appointment</D:contentclass>\n")
                .append("<EX:outlookmessageclass>IPM.Appointment</EX:outlookmessageclass>\n")
                .append("<" + WEBDAV_TAG_APPT_STATE_FLAGS + " dt:dt=\"int\">0</" + WEBDAV_TAG_APPT_STATE_FLAGS + ">\n");

        if (subject != null) {
            subject = StringTools.escapeXml(subject);
            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_SUBJECT)
                    .append(">")
                    .append(subject)
                    .append("</")
                    .append(WEBDAV_TAG_SUBJECT)
                    .append(">\n");
        }

        if (body != null) {
            body = StringTools.escapeXml(body);
            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_BODY)
                    .append(">")
                    .append(body)
                    .append("</")
                    .append(WEBDAV_TAG_BODY)
                    .append(">\n");
        }

        if (location != null) {
            location = StringTools.escapeXml(location);
            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_LOCATION)
                    .append(">")
                    .append(location)
                    .append("</")
                    .append(WEBDAV_TAG_LOCATION)
                    .append(">\n");
        }

        webDavCalendarMsg.append("<")
                .append(WEBDAV_TAG_DATESTART)
                .append(">")
                .append(startDate)
                .append("</")
                .append(WEBDAV_TAG_DATESTART)
                .append(">\n");

        webDavCalendarMsg.append("<")
                .append(WEBDAV_TAG_DATEEND)
                .append(">")
                .append(endDate)
                .append("</")
                .append(WEBDAV_TAG_DATEEND)
                .append(">\n");

        if (allDay != null) {
            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_ALLDAY)
                    .append(">")
                    .append(allDay)
                    .append("</")
                    .append(WEBDAV_TAG_ALLDAY)
                    .append(">\n");
        }

        if (busyStatus != null) {
            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_BUSYSTATUS)
                    .append(">")
                    .append(busyStatus)
                    .append("</")
                    .append(WEBDAV_TAG_BUSYSTATUS)
                    .append(">\n");
        }

        webDavCalendarMsg.append("<")
                .append(WEBDAV_TAG_IMPORTANCE)
                .append(">")
                .append(importance)
                .append("</")
                .append(WEBDAV_TAG_IMPORTANCE)
                .append(">\n");
        //
        // Into WebDav, the WEBDAV_TAG_REMINDER_OFFSET has to be set before
        // WEBDAV_TAG_REMINDER_SET otherwise the reminder results activated
        // with 0 minutes
        //
        webDavCalendarMsg.append("<")
                .append(WEBDAV_TAG_REMINDER_OFFSET)
                .append(">")
                .append((reminderMinutesBeforeStart * 60))
                .append("</")
                .append(WEBDAV_TAG_REMINDER_OFFSET)
                .append(">\n");

        webDavCalendarMsg.append("<")
                .append(WEBDAV_TAG_REMINDER_SET)
                .append(">")
                .append((reminderSet ? "1" : "0"))
                .append("</")
                .append(WEBDAV_TAG_REMINDER_SET)
                .append(">\n");

        if (replyTime != null && replyTime.length() > 0) {
            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_REPLY_TIME)
                    .append(">")
                    .append(replyTime)
                    .append("</")
                    .append(WEBDAV_TAG_REPLY_TIME)
                    .append(">\n");
        }

        if (sensitivity != null) {
            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_SENSITIVITY)
                    .append(">")
                    .append(sensitivity)
                    .append("</")
                    .append(WEBDAV_TAG_SENSITIVITY)
                    .append(">\n");
        }

        if (sequence != null) {
            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_SEQUENCE)
                    .append(">")
                    .append(sequence)
                    .append("</")
                    .append(WEBDAV_TAG_SEQUENCE)
                    .append(">\n");
        }

        if (mileage != null) {
            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_MILEAGE)
                    .append(">")
                    .append(mileage)
                    .append("</")
                    .append(WEBDAV_TAG_MILEAGE)
                    .append(">\n");
        }

        if (!StringTools.isEmpty(categories)) {
            webDavCalendarMsg.append("<EX:keywords-utf8>");

            StringTokenizer st = new StringTokenizer(categories, ";");
            while (st.hasMoreTokens()) {
                webDavCalendarMsg.append("<X:v>")
                        .append(st.nextToken().trim())
                        .append("</X:v>\n");
            }

            webDavCalendarMsg.append("</EX:keywords-utf8>");

        }

        if (recurrence != null && recurrence.length() > 0) {
           webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_RECURRENCE)
                    .append(" dt:dt=\"mv.string\"")
                    .append(">");

            webDavCalendarMsg.append("<X:v>")
                    .append(recurrence)
                    .append("</X:v>\n");

            webDavCalendarMsg.append("</")
                    .append(WEBDAV_TAG_RECURRENCE)
                    .append(">\n");

            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_INSTANCETYPE)
                    .append(" dt:dt=\"int\">1</")
                    .append(WEBDAV_TAG_INSTANCETYPE)
                    .append(">\n");
        } else {
            webDavCalendarMsg.append("<")
                    .append(WEBDAV_TAG_INSTANCETYPE)
                    .append(" dt:dt=\"int\">0</")
                    .append(WEBDAV_TAG_INSTANCETYPE)
                    .append(">\n");
        }

        webDavCalendarMsg
                .append("<").append(WEBDAV_TAG_TIMEZONE_ID).append(">")
                .append(timezoneId)
                .append("</").append(WEBDAV_TAG_TIMEZONE_ID).append(">\n");

        webDavCalendarMsg
                .append(" </D:prop>\n")
                .append("</D:set>\n")
                .append("</D:propertyupdate>");

        webDavHeaderMsg
                = MessageFormat.format(ItemDAO.WEBDAV_HEADER_PROPPATCH,
                new Object[]{
                        "/" +
                                server +
                                "/" +
                                username +
                                "/" +
                                resource,
                        c.getHref()
                }
        );

        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg,
                    credentials,
                    webDavCalendarMsg.toString(),
                    FILE_ENCODING);
        } catch (Exception e) {
            throw new DataAccessException("Error parsing calendar", e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER " +
                    username +
                    " URI " +
                    exchangeFolder +
                    " set Exchange calendar",
                    e);

        }

        try {
            id = XmlParser.getRuidFromResponse(response);
        } catch (XmlParseException e) {
            throw new DataAccessException("Error getting resource id", e);
        }

        c.setId(id);

        return c;

    }

    /**
     * Delete a <i>Calendar</i> from Exchange server.
     *
     * @param calendar
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    public void removeCalendar(Calendar calendar,
                               String username,
                               String credentials,
                               String exchangeFolder)
            throws DataAccessException {

        String webDavCalendarMsg;

        String webDavHeaderMsg;

        String response;

        String server;
        String resource;

        server = getServerFromExchangeFolder(exchangeFolder);
        resource = getResourceFromExchangeFolder(exchangeFolder);

        webDavCalendarMsg = "";


        webDavHeaderMsg =
                MessageFormat.format(ItemDAO.WEBDAV_HEADER_REMOVE,
                        new Object[]{
                                "/" +
                                        server +
                                        "/" +
                                        username +
                                        "/" +
                                        resource,
                                calendar.getHref()

                        }
                );


        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavCalendarMsg, FILE_ENCODING);
        } catch (Exception e) {
            throw new DataAccessException("Error parsing calendar", e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER " +
                    username +
                    " URI " +
                    exchangeFolder +
                    " removing Exchange calendar",
                    e);
        }

    }

    /**
     * get calendar array from Exchange Server
     *
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     * @return array of find calendars
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    public Calendar[] getCalendars(String username,
                                   String credentials,
                                   String[]   ids,
                                   String exchangeFolder)
            throws DataAccessException {

        String clause;

        clause = getClause(ids);

        return getCalendars(username, credentials, clause, exchangeFolder);
    }

    /**
     * get calendar array from Exchange Server
     *
     * @param username
     * @param credentials
     * @param fields
     * @param values
     * @param exchangeFolder
     * @return array of find calendars
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    public Calendar[] getCalendars(String username,
                                   String credentials,
                                   String fields[],
                                   Object values[],
                                   String exchangeFolder) throws DataAccessException {

        String clause;

        clause = getClause(fields, values);

        return getCalendars(username, credentials, clause, exchangeFolder);
    }


    /**
     * get Calendar array with setting fields only for twins
     *
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @param ids
     * @return array of find items
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    public Calendar[] getCalendarsForTwin(String username,
                                          String credentials,
                                          String[] ids,
                                          String exchangeFolder)
            throws DataAccessException {

        ArrayList calendars;
        Calendar c;

        String[]  props;
        String prop;
        String[]  resp;

        String replUid;
        String uri;
        String lastUpdate;

        String dateStart;
        String dateEnd;
        String allDay;
        String location;
        String dtStart;
        String dtEnd;

        boolean isAllDayEvent = false;

        String isFolder;

        String response;

        String webDavBodyMsg;

        String webDavHeaderMsg;

        String server;
        String resource;

        int count = 0;

        String clause;

        server = getServerFromExchangeFolder(exchangeFolder);
        resource = getResourceFromExchangeFolder(exchangeFolder);

        clause = getClause(ids);

        if (clause == null || clause.length() == 0) {
            clause = " WHERE ";
        } else {
            clause = clause + " AND ";
        }

        clause = clause +
                "(\"urn:schemas:calendar:instancetype\" != 2 AND \"urn:schemas:calendar:instancetype\" != 3)";


        webDavBodyMsg =
                MessageFormat.format(WEBDAV_MSG_SELECT_CALENDARS_FOR_TWIN,
                        new Object[]{server, username, resource, clause});

        webDavHeaderMsg =
                MessageFormat.format(WEBDAV_HEADER_SELECT,

                        new Object[]{
                                "/" +
                                        server +
                                        "/" +
                                        username +
                                        "/" +
                                        resource
                        }
                );


        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg,
                    credentials,
                    webDavBodyMsg,
                    FILE_ENCODING);
        } catch (Exception e) {
            throw new DataAccessException("Error getting Exchange server response",
                    e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER " +
                    username +
                    " URI " +
                    exchangeFolder +
                    " reading exchange items",
                    e);
        }

        resp = new String[]{response};

        try {
            props = XmlParser.getXMLTag(resp, TAG_RESPONSE);
        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing item", e);
        }

        calendars = new ArrayList();

        for (int i = 0, l = props.length; i < l; i++) {

            try {

                prop = props[i];

                isFolder = XmlParser.getXMLInitTagValue
                        (prop, TAG_IS_FOLDER);

                replUid = XmlParser.getXMLTagValue
                        (prop, TAG_REPLUID);

                uri = XmlParser.getXMLTagValue
                        (prop, TAG_HREF);

                lastUpdate = XmlParser.getXMLInitTagValue
                        (prop, TAG_LAST_MODIFIED);

                dateStart = XmlParser.getXMLInitTagValue
                        (prop, TAG_DATESTART);

                dateEnd = XmlParser.getXMLInitTagValue
                        (prop, TAG_DATEEND);

                allDay = XmlParser.getXMLInitTagValue
                        (prop, TAG_ALLDAYEVENT);

                location = XmlParser.getXMLTagValue
                        (prop, TAG_LOCATION);

                if (allDay != null) {
                    if ("1".equals(allDay)) {
                        isAllDayEvent = true;
                    } else if ("0".equals(allDay)) {
                        isAllDayEvent = false;
                    }
                }

                if (!isAllDayEvent) {
                    dtStart = DateTools.webDavTagToPDIDate(dateStart, TimeZone.getTimeZone("GMT"));
                    dtEnd = DateTools.webDavTagToPDIDate(dateEnd, TimeZone.getTimeZone("GMT"));

                } else {
                    //
                    // We have to convert the date in yyyy-MM-dd
                    //
                    dtStart = DateTools.convertWebDavDateInDayFormat(dateStart, timezone, true);
                    dtEnd = DateTools.convertWebDavDateInDayFormat(dateEnd, timezone, false);
                }

                c = new Calendar(getIdFromReplUid(replUid), DateTools.webDavTagToDate(lastUpdate));

                c.setHref(XmlParser.getHrefFromUri(uri));

                c.getDtStart().setPropertyValue(dtStart);
                c.getDtEnd().setPropertyValue(dtEnd);

                if (location != null) {
                    c.getLocation().setPropertyValue(location);
                }

                if (PROP_NO_FOLDER.equals(isFolder)) {
                    calendars.add(c);
                    count ++;
                }

            } catch (XmlParseException e) {
                throw new DataAccessException("Error parsing item", e);
            }

        }

        return (Calendar[]) calendars.toArray(new Calendar[count]);
    }

    //---------------------------------------------------------- Private methods

    /**
     * Create a <i>Calendar</i> object from WebDav msg.
     *
     * @param webDavMsg
     * @param webDavMsg
     * @return a newly create <i>Calendar</i> object initialized with the
     *         fields in the result set
     */
    private Calendar[] getCalendarsFromWebDavMsg(String webDavMsg)
            throws DataAccessException {

        ArrayList calendars;
        String  []  resps;
        String  []  msg;

        String response;
        String id;
        String replUid;

        String isFolder;

        int count = 0;

        msg = new String[]{webDavMsg};

        calendars = new ArrayList();

        try {

            resps = XmlParser.getXMLTag(msg, TAG_RESPONSE);

            for (int i = 0, l = resps.length; i < l; i++) {

                response = resps[i];
                replUid = XmlParser.getXMLTagValue(response, TAG_REPLUID);
                id = getIdFromReplUid(replUid);

                isFolder = XmlParser.getXMLInitTagValue(response,
                        TAG_IS_FOLDER);

                if (PROP_NO_FOLDER.equals(isFolder)) {
                    calendars.add(getCalendarFromResponseTag(id, response));
                    count ++;
                }

            }

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing calendar",
                    e);
        }

        return (Calendar[]) calendars.toArray(new Calendar[count]);

    }

    /**
     * Create a <i>Calendar</i> object from WebDav msg.
     *
     * @param id
     * @param msgResponse
     * @return a newly create <i>Calendar</i> object initialized with the
     *         fields in the result set
     */
    private Calendar getCalendarFromResponseTag(String id, String msgResponse)
            throws DataAccessException {

        Calendar c;
        String lastUpdate;

        String uri;

        String dateStart;
        String dateEnd;
        String subject;
        String body;
        String location;
        String allDay;
        String busyStatus;
        String importance;
        String mileage;
        String reminderOffset;
        String reminderSet;
        String replyTime;
        String sensitivity;
        String sequence;

        String[]  categoriesList = null;
        String categories;
        String[]  recurrenceList = null;
        String recurrence;

        try {

            uri = XmlParser.getXMLTagValue
                    (msgResponse, TAG_HREF);

            body = XmlParser.getXMLTagValue
                    (msgResponse, TAG_BODY);

            location = XmlParser.getXMLTagValue
                    (msgResponse, TAG_LOCATION);

            subject = XmlParser.getXMLTagValue
                    (msgResponse, TAG_SUBJECT);

            busyStatus = XmlParser.getXMLTagValue
                    (msgResponse, TAG_BUSY_STATUS);

            dateStart = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_DATESTART);

            dateEnd = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_DATEEND);

            lastUpdate = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_LAST_MODIFIED);

            allDay = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_ALLDAYEVENT);

            importance = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_IMPORTANCE);

            mileage = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_MILEAGE);

            reminderOffset = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_REMINDER_OFFSET);

            reminderSet = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_REMINDER_SET);

            replyTime = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_REPLY_TIME);

            sensitivity = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_SENSITIVITY);

            sequence = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_SEQUENCE);

            categories =
                    XmlParser.getXMLInitTagValue(msgResponse, TAG_CATEGORIES);


            if (categories != null) {
                categoriesList =
                        XmlParser.getXMLTag(new String[]{categories}, "c:v");
            }
            if (categoriesList != null) {
                StringBuffer sb = new StringBuffer();
                for (int i = 0; i < categoriesList.length; i++) {
                    sb.append(categoriesList[i]).append("; ");
                }
                categories = sb.toString();
            }


            recurrence =
                    XmlParser.getXMLInitTagValue(msgResponse, TAG_RECURRENCE);


            if (recurrence != null) {
                recurrence = recurrence.trim();
                recurrenceList =
                        XmlParser.getXMLTag(new String[]{recurrence}, "c:v");
            }
            if (recurrenceList != null) {

                StringBuffer sb = new StringBuffer();

                int l = recurrenceList.length;

                if (l == 1) {

                    sb.append(recurrenceList[0]).append("; ");

                } else {

                    for (int i = 0; i < l; i++) {

                        sb.append(recurrenceList[i]).append("; ");
                    }

                }

                recurrence = sb.toString();
            }

            c = new Calendar(id, DateTools.webDavTagToDate(lastUpdate));

            c.setHref(XmlParser.getHrefFromUri(uri));

            boolean isAllDayEvent = false;

            if (allDay != null) {
                if ("1".equals(allDay) && (recurrence == null || recurrence.length() == 0)) {
                    c.setAllDay(Boolean.TRUE);
                    isAllDayEvent = true;
                } else if ("0".equals(allDay)) {
                    c.setAllDay(Boolean.FALSE);
                }
            }

            String dtStart;
            String dtEnd;

            if (!isAllDayEvent) {
                dtStart = DateTools.webDavTagToPDIDate(dateStart, null);
                dtEnd = DateTools.webDavTagToPDIDate(dateEnd, null);

            } else {
                //
                // We have to convert the date in yyyy-MM-dd
                //
                dtStart = DateTools.convertWebDavDateInDayFormat(dateStart, timezone, true);
                dtEnd = DateTools.convertWebDavDateInDayFormat(dateEnd, timezone, false);
            }

            c.getDtStart().setPropertyValue(dtStart);
            c.getDtEnd().setPropertyValue(dtEnd);

            if (body != null) {
                c.getDescription().setPropertyValue(body);
            }

            if (location != null) {
                c.getLocation().setPropertyValue(location);
            }

            if (subject != null) {
                c.getSummary().setPropertyValue(subject);
            }

            if (importance != null) {
                c.getPriority().setPropertyValue(importance);
            }

            if (mileage != null) {
                c.setMileage(new Integer(mileage));
            }

            if (reminderOffset != null) {
                Reminder reminder = c.getReminder();
                if (reminder == null) {
                    reminder = new Reminder();
                    c.setReminder(reminder);
                }
                int minutes = Integer.parseInt(reminderOffset) / 60;
                reminder.setMinutes(minutes);
            }

            if (reminderSet != null) {
                Reminder reminder = c.getReminder();
                if (reminder == null) {
                    reminder = new Reminder();
                    c.setReminder(reminder);
                }
                if (reminderSet.equals("1")) {
                    reminder.setActive(true);
                } else {
                    reminder.setActive(false);
                }
            }

            if (replyTime != null && replyTime.length() > 0) {
                c.getReplyTime().setPropertyValue(DateTools.webDavTagToPDIDate(replyTime,
                        TimeZone.getTimeZone("GMT")));
            }

            if (sensitivity != null) {
                c.getClassEvent().setPropertyValue(sensitivity);
            }

            if (sequence != null) {
                c.getSequence().setPropertyValue(sequence);
            }

            if (categories != null) {
                c.getCategories().setPropertyValue(categories);
            }

            if (recurrence != null) {
                c.setRecurrencePattern(
                        getRecurrencePattern(recurrence,
                                DateTools.webDavTagToPDIDate(dateStart, TimeZone.getTimeZone("GMT"))));
            }

            short sBusyStatus = (short) 2;

            if ("Free".equalsIgnoreCase(busyStatus)) {
                sBusyStatus = (short) 0;
            } else {
                if ("Tentative".equalsIgnoreCase(busyStatus)) {
                    sBusyStatus = (short) 1;
                } else {
                    if ("Busy".equalsIgnoreCase(busyStatus)) {
                        sBusyStatus = (short) 2;
                    } else {
                        if ("OOF".equalsIgnoreCase(busyStatus)) {
                            sBusyStatus = (short) 3;
                        }
                    }
                }
            }
            c.setBusyStatus(new Short(sBusyStatus));

        } catch (Exception e) {
            throw new DataAccessException("Error parsing calendar item: ", e);
        }

        return c;

    }

    /**
     * Build a clause
     * to webdav select request
     * between ids item array
     *
     * @param fields
     * @param values
     * @return webdav select clause
     */
    private static String getClause(String[] fields, Object[] values) {

        StringBuffer clause = new StringBuffer();

        int l = fields.length;

        if (l == 0) {
            return "";
        }

        clause.append("where (");

        StringBuffer tmp;
        String valueTmp;

        for (int i = 0; i < l; i++) {
            tmp = new StringBuffer();

            valueTmp = "";

            if (TAG_DATESTART.equals(fields[i])) {
                tmp.append("\"").append(FIELD_PREFIX).append("dtstart\"=CAST(\"");

                if (values[i] != null) {
                    valueTmp = DateTools.dateToWebDavTag((java.util.Date) values[i]);
                }
                tmp.append(valueTmp).append("\" as 'dateTime')");
            }

            if (TAG_DATEEND.equals(fields[i])) {
                tmp.append("\"").append(FIELD_PREFIX).append("dtend\"=CAST(\"");

                if (values[i] != null) {
                    valueTmp = DateTools.dateToWebDavTag((java.util.Date) values[i]);
                }

                tmp.append(valueTmp).append("\" as 'dateTime')");
            }

            if (TAG_LOCATION.equals(fields[i])) {
                tmp.append("\"").append(FIELD_PREFIX).append("location\"");

                if (values[i] != null) {
                    tmp.append("='");
                    valueTmp = String.valueOf(values[i]);
                    valueTmp = valueTmp.replaceAll("'", "''");
                    valueTmp = StringTools.escapeXml(valueTmp);
                    tmp.append(valueTmp).append("'");
                } else {
                    tmp.append(" is null ");
                }

            }

            if (i != 0 && tmp.length() != 0) {
                clause.append(" and ").append(tmp);
            } else if (tmp.length() != 0) {
                clause.append(tmp);
            }

        }

        clause.append(")\r\n");

        return clause.toString();

    }


    /**
     * get calendars from Exchange Server with the given clause
     *
     * @param username
     * @param credentials
     * @param clause
     * @param exchangeFolder
     * @return array of find calendars
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    private Calendar[] getCalendars(String username,
                                    String credentials,
                                    String clause,
                                    String exchangeFolder)
            throws DataAccessException {

        Calendar[] exchangeCalendars;

        String response;
        String webDavCalendarMsg;
        String webDavHeaderMsg;

        String server;
        String resource;

        server = getServerFromExchangeFolder(exchangeFolder);
        resource = getResourceFromExchangeFolder(exchangeFolder);

        if (clause == null || clause.length() == 0) {
            clause = " WHERE ";
        } else {
            clause = clause + " AND ";
        }

        clause = clause +
                "(\"urn:schemas:calendar:instancetype\" != 2 AND \"urn:schemas:calendar:instancetype\" != 3)";

        webDavCalendarMsg =
                MessageFormat.format(WEBDAV_MSG_SELECT_CALENDARS,
                        new Object[]{server, username, resource,
                                clause});

        webDavHeaderMsg =
                MessageFormat.format(ItemDAO.WEBDAV_HEADER_SELECT,
                        new Object[]{
                                "/" +
                                        server +
                                        "/" +
                                        username +
                                        "/" +
                                        resource
                        }
                );

        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials,
                    webDavCalendarMsg, FILE_ENCODING);
        } catch (Exception e) {
            throw new DataAccessException(
                    "Error getting Exchange server response",
                    e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER " +
                    username +
                    " URI " +
                    exchangeFolder +
                    " getting exchange calendar",
                    e);

        }

        exchangeCalendars = getCalendarsFromWebDavMsg(response);

        return exchangeCalendars;

    }

    /**
     * <p> Maps an Exchange rrule to a RecurrencePattern. </p>
     *
     * @param recurrence       the input rrule
     * @param startDatePattern the starting date of the recurrence defined by rrule
     * @return the RecurrencePattern corresponding to the input rrule and start date
     * @throws Exception if an error occurs
     */
    private RecurrencePattern getRecurrencePattern(String recurrence,
                                                   String startDatePattern)
            throws Exception {

        RecurrencePattern rp = null;

        short recurrenceType = 0;
        short dayOfWeekMask = 0;
        short dayOfMonth = 0;
        short monthOfYear = 0;
        short instance = 0;

        int interval = 0;
        int occurrences = 0;

        String rt;
        String value;
        String endDatePattern = null;

        boolean noEndDate = true;

        rt = recurrence.substring(recurrence.indexOf("FREQ=") + "FREQ=".length(),
                recurrence.indexOf(";")).trim();

        //
        // start date pattern
        //
        if (recurrence.indexOf("UNTIL") != -1) {
            noEndDate = false;
            int start = recurrence.indexOf("UNTIL=") + "UNTIL=".length();
            int end = recurrence.indexOf(";", start);
            endDatePattern = recurrence.substring(start, end).trim();

        }

        //
        // interval
        //
        if (recurrence.indexOf("INTERVAL") != -1) {
            int start = recurrence.indexOf("INTERVAL=") + "INTERVAL=".length();
            int end = recurrence.indexOf(";", start);
            value = recurrence.substring(start, end).trim();
            interval = Integer.parseInt(value);

        }

        if (recurrence.indexOf("COUNT") != -1) {
            noEndDate = false;
            int start = recurrence.indexOf("COUNT=") + "COUNT=".length();
            int end = recurrence.indexOf(";", start);
            value = recurrence.substring(start, end).trim();
            occurrences = Integer.parseInt(value);
        }

        if ("DAILY".equals(rt)) {
            //
            // daily
            //
            if (recurrence.indexOf(EVERY_WEEKDAY) != -1) {
                recurrenceType = RecurrencePattern.TYPE_WEEKLY;
                dayOfWeekMask = getDayOfWeekMask(EVERY_WEEKDAY);
            } else {
                recurrenceType = RecurrencePattern.TYPE_DAYLY;
            }
        } else if ("WEEKLY".equals(rt)) {
            //
            // weekly
            //
            int start = recurrence.indexOf("BYDAY=") + "BYDAY=".length();
            int end = recurrence.indexOf(";", start);
            value = recurrence.substring(start, end).trim();
            dayOfWeekMask = getDayOfWeekMask(value);
            recurrenceType = RecurrencePattern.TYPE_WEEKLY;
        } else if ("MONTHLY".equals(rt)) {
            if (recurrence.indexOf("BYDAY") != - 1) {
                //
                // montly nt
                //
                int start = recurrence.indexOf("BYDAY=") + "BYDAY=".length();
                int end = recurrence.indexOf(";", start);
                value = recurrence.substring(start, end).trim();
                if (value.length() > 1 && value.substring(0, 1).matches("[1-5]")) {
                    instance = Short.parseShort(value.substring(0, 1));
                    dayOfWeekMask = getDayOfWeekMask(value.substring(1).trim());
                } else {
                    //
                    // last
                    //
                    instance = 5;
                    dayOfWeekMask = getDayOfWeekMask(value);
                }

                recurrenceType = RecurrencePattern.TYPE_MONTH_NTH;

            } else {
                //
                // montly
                //
                int start = recurrence.indexOf("BYMONTHDAY=") + "BYMONTHDAY=".length();
                int end = recurrence.indexOf(";", start);
                value = recurrence.substring(start, end).trim();
                dayOfMonth = Short.parseShort(value);
                recurrenceType = RecurrencePattern.TYPE_MONTHLY;
            }

        } else if ("YEARLY".equals(rt)) {

            int start = recurrence.indexOf("BYMONTH=") + "BYMONTH=".length();
            int end = recurrence.indexOf(";", start);
            value = recurrence.substring(start, end).trim();
            monthOfYear = Short.parseShort(value);

            if (recurrence.indexOf("BYDAY") != - 1) {
                //
                // yearly nt
                //
                start = recurrence.indexOf("BYDAY=") + "BYDAY=".length();
                end = recurrence.indexOf(";", start);
                value = recurrence.substring(start, end).trim();
                if (value.length() > 1 && value.substring(0, 1).matches("[1-5]")) {
                    instance = Short.parseShort(value.substring(0, 1));
                    dayOfWeekMask = getDayOfWeekMask(value.substring(1).trim());
                } else {
                    //
                    // last
                    //
                    instance = 5;
                    dayOfWeekMask = getDayOfWeekMask(value);
                }

                recurrenceType = RecurrencePattern.TYPE_YEAR_NTH;
            } else {
                //
                // yearly
                //
                start = recurrence.indexOf("BYMONTHDAY=") + "BYMONTHDAY=".length();
                end = recurrence.indexOf(";", start);
                value = recurrence.substring(start, end).trim();
                dayOfMonth = Short.parseShort(value);

                recurrenceType = RecurrencePattern.TYPE_YEARLY;
            }

        }

        switch (recurrenceType) {
            case RecurrencePattern.TYPE_DAYLY:

                if (!noEndDate) {
                    rp = RecurrencePattern.getDailyRecurrencePattern(interval,
                            startDatePattern,
                            endDatePattern,
                            noEndDate,
                            occurrences);
                } else {
                    rp = RecurrencePattern.getDailyRecurrencePattern(interval,
                            startDatePattern,
                            noEndDate);
                }
                break;
            case RecurrencePattern.TYPE_WEEKLY:
                if (!noEndDate) {
                    rp = RecurrencePattern.getWeeklyRecurrencePattern(interval,
                            dayOfWeekMask,
                            startDatePattern,
                            endDatePattern,
                            noEndDate,
                            occurrences);
                } else {
                    rp = RecurrencePattern.getWeeklyRecurrencePattern(interval,
                            dayOfWeekMask,
                            startDatePattern,
                            noEndDate);
                }
                break;
            case RecurrencePattern.TYPE_MONTHLY:
                if (!noEndDate) {
                    rp = RecurrencePattern.getMonthlyRecurrencePattern(interval,
                            dayOfMonth,
                            startDatePattern,
                            endDatePattern,
                            noEndDate,
                            occurrences);

                } else {
                    rp = RecurrencePattern.getMonthlyRecurrencePattern(interval,
                            dayOfMonth,
                            startDatePattern,
                            noEndDate);
                }
                break;
            case RecurrencePattern.TYPE_MONTH_NTH:
                if (!noEndDate) {
                    rp = RecurrencePattern.getMonthNthRecurrencePattern(interval,
                            dayOfWeekMask,
                            instance,
                            startDatePattern,
                            endDatePattern,
                            noEndDate,
                            occurrences);
                } else {
                    rp = RecurrencePattern.getMonthNthRecurrencePattern(interval,
                            dayOfWeekMask,
                            instance,
                            startDatePattern,
                            noEndDate);
                }
                break;
            case RecurrencePattern.TYPE_YEARLY:

                if (!noEndDate) {
                    rp = RecurrencePattern.getYearlyRecurrencePattern(interval,
                            dayOfMonth,
                            monthOfYear,
                            startDatePattern,
                            endDatePattern,
                            noEndDate,
                            occurrences);

                } else {
                    rp = RecurrencePattern.getYearlyRecurrencePattern(interval,
                            dayOfMonth,
                            monthOfYear,
                            startDatePattern,
                            noEndDate);
                }
                break;
            case RecurrencePattern.TYPE_YEAR_NTH:
                if (!noEndDate) {
                    rp = RecurrencePattern.getYearNthRecurrencePattern(interval,
                            dayOfWeekMask,
                            monthOfYear,
                            instance,
                            startDatePattern,
                            endDatePattern,
                            noEndDate,
                            occurrences);
                } else {
                    rp = RecurrencePattern.getYearNthRecurrencePattern(interval,
                            dayOfWeekMask,
                            monthOfYear,
                            instance,
                            startDatePattern,
                            noEndDate);
                }

        }

        if (rp == null) {
            return null;
        }

        return rp;
    }

    /**
     * <p> Maps a RecurrencePattern to the rrule accepted by Exchange. </p>
     *
     * @param rp the input RecurrencePattern
     * @return rrule corresponding to the input RecurrencePattern
     * @throws Exception if an error occurs
     */
    private String getExchangeRecurrence(RecurrencePattern rp, String startDate)
            throws Exception {

        String exr = null;

        String interval;
        String endPatternDate;
        String dayOfWeekMask;
        String dayOfMonth;
        String monthOfYear;

        boolean noEndDate;

        short instance;

        if (rp.getInterval() > 0) {
            interval = ";INTERVAL=" + rp.getInterval();
        } else {
            interval = ";INTERVAL=1";
        }


        endPatternDate = rp.getEndDatePattern();


        noEndDate = rp.isNoEndDate();

        if (!noEndDate) {
            if (rp.getOccurrences() > 0) {
                endPatternDate = ";COUNT=" + rp.getOccurrences();
            } else {
                endPatternDate = ";UNTIL=" + endPatternDate;
            }
        } else {
            endPatternDate = "";
        }

        switch (rp.getTypeId()) {
            case RecurrencePattern.TYPE_DAYLY:
                dayOfWeekMask = getDayOfWeekMask(rp.getDayOfWeekMask());
                if (dayOfWeekMask != null && dayOfWeekMask.length() > 0) {
                    dayOfWeekMask = ";BYDAY=" + dayOfWeekMask;
                } else {
                    dayOfWeekMask = "";
                }
                exr = "FREQ=DAILY" +
                        endPatternDate +
                        interval +
                        dayOfWeekMask +
                        ";WKST=SU";

                break;
            case RecurrencePattern.TYPE_WEEKLY:

                dayOfWeekMask = getDayOfWeekMask(rp.getDayOfWeekMask());

                if (dayOfWeekMask != null && dayOfWeekMask.length() > 0) {
                    dayOfWeekMask = ";BYDAY=" + dayOfWeekMask;
                } else {
                    dayOfWeekMask = "";
                }

                exr = "FREQ=WEEKLY" +
                        endPatternDate +
                        interval +
                        dayOfWeekMask +
                        ";WKST=SU";
                break;
            case RecurrencePattern.TYPE_MONTHLY:

                if (rp.getDayOfMonth() > 0) {
                    dayOfMonth = ";BYMONTHDAY=" + rp.getDayOfMonth();
                } else {
                    short dayOfMonthVal = 0;

                    if (rp.getStartDatePattern() != null) {
                        dayOfMonthVal = getDayOfTheMonth(rp.getStartDatePattern(),
                                "yyyyMMdd'T'HHmmss'Z'",
                                null,
                                timezone);
                    } else {
                        if (startDate != null) {
                            dayOfMonthVal = getDayOfTheMonth(startDate,
                                    "yyyyMMdd'T'HHmmss'Z'",
                                    null,
                                    timezone);
                        }
                    }

                    dayOfMonth = ";BYMONTHDAY=" + dayOfMonthVal;
                }

                exr = "FREQ=MONTHLY" +
                        endPatternDate +
                        interval +
                        dayOfMonth +
                        ";WKST=SU";

                break;
            case RecurrencePattern.TYPE_MONTH_NTH:

                String instanceDayOfWeekMask;

                instance = rp.getInstance();

                dayOfWeekMask = getDayOfWeekMask(rp.getDayOfWeekMask());

                if (dayOfWeekMask == null) {
                    dayOfWeekMask = "";
                }
                instanceDayOfWeekMask = ";BYDAY=" + instance + " " + dayOfWeekMask;

                exr = "FREQ=MONTHLY" +
                        endPatternDate +
                        interval +
                        instanceDayOfWeekMask +
                        ";WKST=SU";

                break;
            case RecurrencePattern.TYPE_YEARLY:

                dayOfMonth = ";BYMONTHDAY=" + rp.getDayOfMonth();
                monthOfYear = ";BYMONTH=" + rp.getMonthOfYear();

                exr = "FREQ=YEARLY" +
                        endPatternDate +
                        dayOfMonth +
                        monthOfYear +
                        ";WKST=SU";

                break;

            case RecurrencePattern.TYPE_YEAR_NTH:

                instance = rp.getInstance();

                dayOfWeekMask = getDayOfWeekMask(rp.getDayOfWeekMask());

                if (dayOfWeekMask == null) {
                    dayOfWeekMask = "";
                }
                instanceDayOfWeekMask = ";BYDAY=" + instance + " " + dayOfWeekMask;
                monthOfYear = ";BYMONTH=" + rp.getMonthOfYear();

                exr = "FREQ=YEARLY" +
                        endPatternDate +
                        interval +
                        instanceDayOfWeekMask +
                        monthOfYear +
                        ";WKST=SU";

                break;


        }

        return exr;


    }

    /**
     * <p> Returns a numerical representation accepted by SIF-C standard of the input day of the week mask. </p>
     *
     * @param value the input string day of the week mask
     * @return numerical representation accepted by SIF-C of the input day of the week mask
     */
    private short getDayOfWeekMask(String value) {

        short dayOfWeekMask = 0;
        String day;

        StringTokenizer st = new StringTokenizer(value, ",");

        while (st.hasMoreTokens()) {

            day = st.nextToken().trim();

            if ("SU".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_SUNDAY);
            } else if ("MO".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_MONDAY);
            } else if ("TU".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_TUESDAY);
            } else if ("WE".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_WEDNSDAY);
            } else if ("TH".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_THURSDAY);
            } else if ("FR".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_FRIDAY);
            } else if ("SA".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_SATURDAY);
            }

        }

        return dayOfWeekMask;

    }

    /**
     * <p> Returns a string representation accepted by Exchange of the input day of the week mask. </p>
     *
     * @param value the input mumerical day of the week mask
     * @return string representation accepted by Exchange of the input day of the week mask
     */
    private String getDayOfWeekMask(short value) {

        String dayOfWeekMask = "";

        if ((value & RecurrencePattern.DAY_OF_WEEK_SUNDAY) != 0) {
            dayOfWeekMask = "SU";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_MONDAY) != 0) {

            if (dayOfWeekMask.length() > 0) {
                dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "MO";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_TUESDAY) != 0) {

            if (dayOfWeekMask.length() > 0) {
                dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "TU";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_WEDNSDAY) != 0) {

            if (dayOfWeekMask.length() > 0) {
                dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "WE";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_THURSDAY) != 0) {

            if (dayOfWeekMask.length() > 0) {
                dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "TH";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_FRIDAY) != 0) {

            if (dayOfWeekMask.length() > 0) {
                dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "FR";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_SATURDAY) != 0) {

            if (dayOfWeekMask.length() > 0) {
                dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "SA";
        }

        return dayOfWeekMask;
    }

    /**
     * <p> Map the input timezone to the corresponding Exchange accepted timezone id. </p>
     *
     * @param tz the input timezone
     * @return the timezone id corresponding to the input timezone
     */
    private static int getTimezoneId(TimeZone tz) {
        HashMap idsToKeyWordsMap, offsetsToIdsMap;
        int tzRawOffset;
        String tzDesc;
        String[] ids, keyWords;

        if (tz == null) {
            return 0;
        }

        idsToKeyWordsMap = new HashMap();

        idsToKeyWordsMap.put("1", new String[]{"Greenwich"});
        idsToKeyWordsMap.put("2", new String[]{"Dublin", "Edinburgh", "Lisbon", "London"});
        idsToKeyWordsMap.put("3", new String[]{"Brussels", "Copenhagen", "Madrid", "Paris", "Vilnius"});
        idsToKeyWordsMap.put("4", new String[]{"Amsterdam", "Berlin", "Bern", "Rome", "Stockholm", "Vienna"});
        idsToKeyWordsMap.put("5", new String[]{"Eastern Europe"});
        idsToKeyWordsMap.put("6", new String[]{"Belgrade", "Pozsony", "Budapest", "Ljubljana", "Prague"});
        idsToKeyWordsMap.put("7", new String[]{"Athens", "Instanbul", "Minsk"});
        idsToKeyWordsMap.put("8", new String[]{"Brasilia"});
        idsToKeyWordsMap.put("9", new String[]{"Atlantic time"});
        idsToKeyWordsMap.put("10", new String[]{"Eastern time"});
        idsToKeyWordsMap.put("11", new String[]{"Central time"});
        idsToKeyWordsMap.put("12", new String[]{"Mountain time"});
        idsToKeyWordsMap.put("13", new String[]{"Pacific time"});
        idsToKeyWordsMap.put("14", new String[]{"Alaska"});
        idsToKeyWordsMap.put("15", new String[]{"Hawaii"});
        idsToKeyWordsMap.put("16", new String[]{"Midway Island", "Samoa"});
        idsToKeyWordsMap.put("17", new String[]{"Auckland", "Wellington"});
        idsToKeyWordsMap.put("18", new String[]{"Brisbane"});
        idsToKeyWordsMap.put("19", new String[]{"Adelaide"});
        idsToKeyWordsMap.put("20", new String[]{"Osaka", "Sapporo", "Tokyo"});
        idsToKeyWordsMap.put("21", new String[]{"Hong Kong SAR"});
        idsToKeyWordsMap.put("22", new String[]{"Bangkok", "Hanoi", "Jakarta"});
        idsToKeyWordsMap.put("23", new String[]{"Mumbai", "Kolkata", "Chennai", "New Delhi"});
        idsToKeyWordsMap.put("24", new String[]{"Abu Dhabi", "Muscat"});
        idsToKeyWordsMap.put("25", new String[]{"Tehran"});
        idsToKeyWordsMap.put("26", new String[]{"Baghdad, Kuwait, Riyadh"});
        idsToKeyWordsMap.put("27", new String[]{"Israel"});
        idsToKeyWordsMap.put("28", new String[]{"Newfoundland"});
        idsToKeyWordsMap.put("29", new String[]{"Azores", "Cape Verde Islands"});
        idsToKeyWordsMap.put("30", new String[]{"Mid Atlantic"});
        idsToKeyWordsMap.put("31", new String[]{"Casablanca", "Monrovia"});
        idsToKeyWordsMap.put("32", new String[]{"Buenos Aires", "Georgetown"});
        idsToKeyWordsMap.put("33", new String[]{"Caracas", "La Paz"});
        idsToKeyWordsMap.put("34", new String[]{"Indiana"});
        idsToKeyWordsMap.put("35", new String[]{"Bogota", "Lima", "Quito"});
        idsToKeyWordsMap.put("36", new String[]{"Saskatchewan"});
        idsToKeyWordsMap.put("37", new String[]{"Mexico City", "Tegucigalpa"});
        idsToKeyWordsMap.put("38", new String[]{"Arizona"});
        idsToKeyWordsMap.put("39", new String[]{"Eniwetok", "Kwajalein"});
        idsToKeyWordsMap.put("40", new String[]{"Fiji Islands", "Kamchatka", "Marshall Islands"});
        idsToKeyWordsMap.put("41", new String[]{"Magadan", "Solomon Islands", "New Caledonia"});
        idsToKeyWordsMap.put("42", new String[]{"Hobart"});
        idsToKeyWordsMap.put("43", new String[]{"Guam, Port Moresby"});
        idsToKeyWordsMap.put("44", new String[]{"Darwin"});
        idsToKeyWordsMap.put("45", new String[]{"Beijing", "Chongqing", "Urumqi"});
        idsToKeyWordsMap.put("46", new String[]{"Akmola", "Almaty", "Dhaka"});
        idsToKeyWordsMap.put("47", new String[]{"Islamabad", "Karachi", "Tashkent"});
        idsToKeyWordsMap.put("48", new String[]{"Kabul"});
        idsToKeyWordsMap.put("49", new String[]{"Cairo"});
        idsToKeyWordsMap.put("50", new String[]{"Harare", "Pretoria"});
        idsToKeyWordsMap.put("51", new String[]{"Moscow", "St. Petersburg", "Volgograd"});

        offsetsToIdsMap = new HashMap();

        offsetsToIdsMap.put("-12", new String[]{"39"});
        offsetsToIdsMap.put("-11", new String[]{"16"});
        offsetsToIdsMap.put("-10", new String[]{"15"});
        offsetsToIdsMap.put("-9", new String[]{"14"});
        offsetsToIdsMap.put("-8", new String[]{"13"});
        offsetsToIdsMap.put("-7", new String[]{"12", "38"});
        offsetsToIdsMap.put("-6", new String[]{"11", "36", "37"});
        offsetsToIdsMap.put("-5", new String[]{"10", "34", "35"});
        offsetsToIdsMap.put("-4", new String[]{"9", "33"});
        offsetsToIdsMap.put("-3", new String[]{"8", "28", "32"});
        offsetsToIdsMap.put("-2", new String[]{"30"});
        offsetsToIdsMap.put("-1", new String[]{"29"});
        offsetsToIdsMap.put("0", new String[]{"0", "1", "2", "31"});
        offsetsToIdsMap.put("1", new String[]{"3", "4", "6"});
        offsetsToIdsMap.put("2", new String[]{"5", "27", "49", "50"});
        offsetsToIdsMap.put("3", new String[]{"26", "25", "51"});
        offsetsToIdsMap.put("4", new String[]{"24", "48"});
        offsetsToIdsMap.put("5", new String[]{"47", "23"});
        offsetsToIdsMap.put("6", new String[]{"46"});
        offsetsToIdsMap.put("7", new String[]{"22"});
        offsetsToIdsMap.put("8", new String[]{"45", "21"});
        offsetsToIdsMap.put("9", new String[]{"20", "19", "44"});
        offsetsToIdsMap.put("10", new String[]{"18", "42", "43"});
        offsetsToIdsMap.put("11", new String[]{"41"});
        offsetsToIdsMap.put("12", new String[]{"17", "40"});


        tzRawOffset = tz.getRawOffset() / 3600000;
        tzDesc = tz.getDisplayName();

        ids = (String[]) offsetsToIdsMap.get(String.valueOf(tzRawOffset));
        if (ids != null) {
            for (int i = 1; i < ids.length; i++) {
                keyWords = (String[]) idsToKeyWordsMap.get(ids[i]);
                for (int j = 0; j < keyWords.length; j++) {
                    if (tzDesc.indexOf(keyWords[j]) > -1) {
                        return Integer.parseInt(ids[i]);
                    }
                }
            }
            return Integer.parseInt(ids[0]);
        }

        return 0;
    }

    /**
     * <p> Get the day of the month from a string representation of a date, and mappes it to a different timezone.</p>
     *
     * @param strDate     the string representation of the datre
     * @param pattern     the date pattern
     * @param inTimeZone  the timezone of the input
     * @param outTimeZone the timezone of the output
     * @return the day of the month
     * @throws ParseException is strDate doesn't conform to pattern
     */
    public static short getDayOfTheMonth(String strDate, String pattern, TimeZone inTimeZone, TimeZone outTimeZone)
            throws ParseException {

        Date date;
        DateFormat dateFormat;
        GregorianCalendar calendar;

        if (strDate == null || pattern == null) {
            return 0;
        }

        if (inTimeZone == null) {
            inTimeZone = TimeZone.getTimeZone("GMT");
        }

        if (outTimeZone == null) {
            outTimeZone = TimeZone.getTimeZone("GMT");
        }

        dateFormat = new SimpleDateFormat(pattern);
        dateFormat.setTimeZone(inTimeZone);
        date = dateFormat.parse(strDate);

        calendar = new GregorianCalendar(outTimeZone);
        calendar.setTime(date);
        return (short) calendar.get(GregorianCalendar.DAY_OF_MONTH);
    }
}

