/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.items.common.dao;

import java.io.IOException;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Timestamp;

import java.text.MessageFormat;

import java.util.ArrayList;
import java.util.Date;
import java.util.TimeZone;

import com.funambol.framework.server.store.NotFoundException;

import com.funambol.exchange.AuthenticationException;
import com.funambol.exchange.DataAccessException;
import com.funambol.exchange.ExchangeAccessException;
import com.funambol.exchange.httptransport.WebDavHttpTransport;
import com.funambol.exchange.items.common.model.Item;
import com.funambol.exchange.xml.XmlParseException;
import com.funambol.exchange.util.DataAccess;

/*
 * This class implement methods to access items common data
 * in database source and exchange datastore
 *
 * @version $Id: ItemDAO.java,v 1.1 2006/08/08 07:48:08 gibi Exp $
 *
 **/
public class ItemDAO {

    //---------------------------------------------------------------- Constants

    // Define a common initial size for the StringBuffer.
    protected static final int INITIAL_SIZE_SBUFFER = 2048; // 2K

    protected static final String FILE_ENCODING = "UTF-8";

    protected static final String WEBDAV_HEADER_PROPPATCH                  =
    "PROPPATCH {0}/{1} HTTP/1.0\r\n";

    protected static final String WEBDAV_HEADER_REMOVE                     =
    "DELETE {0}/{1} HTTP/1.0\r\n";

    protected static final String WEBDAV_HEADER_SELECT                     =
    "SEARCH {0} HTTP/1.0\r\n"                                              +
    "Depth: 1,noroot\r\n";

    private static final String WEBDAV_HEADER_PROPFIND                     =
    "PROPFIND {0} HTTP/1.0\r\n"                                            +
    "Depth: 1,noroot\r\n";

    private static final String WEBDAV_HEADER_PROPFIND_ACCESS_STATUS       =
    "PROPFIND {0} HTTP/1.0\r\n"                                            +
    "Depth: 0\r\n";

    private static final String WEBDAV_MSG_ACCESS_STATUS                   =
    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                        +
    "<D:propfind xmlns:D=\"DAV:\">\n"                                      +
    "<D:allprop/>\n"                                                       +
    "</D:propfind>\n"                                                      ;

    private static final String CLAUSE_REPLID_FIELD
        = "\"http://schemas.microsoft.com/repl/repl-uid\""          ;

    private static final String CONTENT_REPLID_START   = "rid:"     ;

    private static final String SQL_SELECT_ITEMS_BY_PRINCIPAL_BY_SOURCE    =
    "select id, href, last_modified, twin_crc "                            +
    "from fnbl_exch_cache "                                          +
    "where "                                                               +
    "uri = ? and principal = ?"                                            ;

    private static final String SQL_SELECT_HREF_BY_BY_ID_PRINCIPAL_BY_SOURCE =
    "select href "                                                         +
    "from fnbl_exch_cache "                                          +
    "where "                                                               +
    "uri = ? and principal = ? and id = ?"                                 ;

    private static final String SQL_GET_TWIN                               =
     "select id, href, last_modified, twin_crc "                           +
     "from fnbl_exch_cache "                                         +
     "where "                                                              +
     "uri = ? and principal = ? and twin_crc = ?"                          ;

    private static final String SQL_ADD_LOCAL_ITEM                         =
    "insert into fnbl_exch_cache "                                   +
    "(id, href, last_modified, uri, principal, twin_crc) "                 +
    "values(?, ?, ?, ?, ?, ?)"                                             ;

    private static final String SQL_UPDATE_LOCAL_ITEM                      =
    "update fnbl_exch_cache "                                        +
    "set last_modified = ? , twin_crc = ? "                                +
    "where id = ? and uri = ? and principal = ?"                           ;

    private static final String SQL_DELETE_LOCAL_ITEM                      =
    "delete from fnbl_exch_cache "                                   +
    "where id = ? and uri = ? and principal = ? "                          ;

    protected static final String TAG_HREF           = "href"              ;
    protected static final String TAG_CREATION_DATE  = "creationdate"      ;
    private   static final String TAG_REPLUID        = "repluid"           ;
    private   static final String TAG_LAST_MODIFIED  = "getlastmodified"   ;
    protected
              static final String TAG_IS_FOLDER      = "isfolder"          ;
    private   static final String TAG_PROP           = "a:prop"            ;

    protected
        static final String PROP_NO_FOLDER           = "0"                 ;

    protected static final String NEW_LINE           = "\r\n"              ;

    private static final int
        STATUS_ACCESS_DANIED                         = 401                 ;
    private static final int
        STATUS_OK_LOAD_CHANGE                        = 207                 ;
    private static final int
        STATUS_OK_DELETE                             = 200                 ;
    private static final int
        STATUS_NOT_FOUND                             = 404                 ;

    public static final String TAG_RESPONSE           = "a:response"       ;

    //------------------------------------------------------------- Private data

    private DataAccess           da                  = null ;
    private WebDavHttpTransport  webDavHttp          = null ;

    //-------------------------------------------------------------- Costructors

    public ItemDAO ()
        throws DataAccessException {

        this.da = new DataAccess ()  ;

    }

    public ItemDAO (String  host ,
                    int     port ,
                    String  exchangeServer)
    throws DataAccessException {

        this.da          = new DataAccess ()  ;
        this.webDavHttp  = new WebDavHttpTransport(host, port, exchangeServer);

    }

    //----------------------------------------------------------- Public methods
    /**
     * get local items
     *
     * @param sourceURI
     * @param principal
     *
     * @return array of find local items
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public Item[] getLocalItems (String sourceURI, long principal)
    throws DataAccessException {

         Connection         connection     = null ;
         PreparedStatement  ps             = null ;
         ResultSet          rs             = null ;

         ArrayList          rows           = null ;

         Item[]             items          = null ;
         Item               item           = null ;

         int                count          = 0    ;

         try {

             connection = this.da.getConnection();

             rows = new ArrayList();

             ps = connection.prepareStatement(
                    SQL_SELECT_ITEMS_BY_PRINCIPAL_BY_SOURCE);

             ps.setString (1, sourceURI ) ;
             ps.setLong   (2, principal ) ;

             rs = ps.executeQuery();

             while(rs.next()) {
                 rows.add(new Item(rs.getString(1),
                                 rs.getString(2),
                                 new Date(rs.getTimestamp(3).getTime()),
                                 rs.getLong(4)));
                 ++count;
             }

             items = (Item[]) rows.toArray(new Item[count]);

         } catch (Exception e) {
             throw new DataAccessException("Error getting local items", e);
         } finally {
             try {
                 this.da.cleanUp(connection, ps, rs);
             } catch (Exception e)  {
                 throw new DataAccessException("Error getting local items", e);
             }
         }

         return items;

    }

     /**
      * get href
      *
      * @param id
      * @param principal
      * @param sourceURI
      *
      * @return href
      *
      * @throws com.funambol.exchange.DataAccessException
      **/
     public String getHref (String id, long principal, String sourceURI)
     throws DataAccessException {

          Connection         connection     = null ;
          PreparedStatement  ps             = null ;
          ResultSet          rs             = null ;

          String             href           = null ;

          try {

              connection = this.da.getConnection();

              ps = connection.prepareStatement(
                     SQL_SELECT_HREF_BY_BY_ID_PRINCIPAL_BY_SOURCE);

              ps.setString  (1, sourceURI ) ;
              ps.setLong    (2, principal ) ;
              ps.setString  (3, id        ) ;


              rs = ps.executeQuery();

              while(rs.next()) {
                  href = rs.getString(1);
              }

          } catch (Exception e) {
              throw new DataAccessException("Error getting href", e);
          } finally {
              try {
                  this.da.cleanUp(connection, ps, rs);
              } catch (Exception e)  {
                  throw new DataAccessException("Error getting href", e);
              }
          }

         return href;

     }

    /**
     * add local items
     *
     * @param items
     * @param sourceURI
     * @param principal
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public void addItems(Item[] items     ,
                         String sourceURI ,
                         long   principal )
    throws DataAccessException {

         Connection         connection    = null ;
         PreparedStatement  ps            = null ;

         String             id            = null ;
         String             href          = null ;
         Date               lastModified  = null ;

         try {

            connection = this.da.getConnection();

            for (int i=0, l = items.length; i < l; i++) {

                ps = connection.prepareStatement(SQL_ADD_LOCAL_ITEM);

                id            = items[i].getId()           ;
                href          = items[i].getHref()         ;
                lastModified  = items[i].getLastModified() ;

                ps.setString    (1, id                                    ) ;
                ps.setString    (2, href                                  ) ;
                ps.setTimestamp (3, new Timestamp(lastModified.getTime()) ) ;

                ps.setString    (4, sourceURI                             ) ;
                ps.setLong      (5, principal                             ) ;

                ps.setLong      (6, items[i].getCRC()                     ) ;

                ps.executeUpdate();

                ps.close() ;
                ps = null  ;

            }

         } catch (Exception e) {
              throw new DataAccessException("Error adding local items", e);
         } finally {
             try {
                 this.da.cleanUp(connection, ps, null);
             } catch (Exception e)  {
                  throw new DataAccessException("Error adding local items", e);
             }
         }

    }

     /**
      * add local items
      *
      * @param items
      * @param sourceURI
      * @param principal
      *
      * @throws com.funambol.exchange.DataAccessException
      **/
     public void updateItems(Item[] items     ,
                             String sourceURI ,
                             long   principal)
     throws DataAccessException {

          Connection         connection    = null ;
          PreparedStatement  ps            = null ;

          String             id            = null;
          Date               lastModified  = null;

          try {

             connection = this.da.getConnection();

             for (int i=0, l = items.length; i < l; i++) {

                 ps = connection.prepareStatement(SQL_UPDATE_LOCAL_ITEM);

                 id            = items[i].getId()           ;
                 lastModified  = items[i].getLastModified() ;

                 ps.setTimestamp (1, new Timestamp(lastModified.getTime()) ) ;
                 ps.setLong      (2, items[i].getCRC()                     ) ;
                 ps.setString    (3, id                                    ) ;
                 ps.setString    (4, sourceURI                             ) ;
                 ps.setLong      (5, principal                             ) ;

                 ps.executeUpdate();

                 ps.close() ;
                 ps = null  ;

             }

          } catch (Exception e) {
              throw new DataAccessException("Error updating local items", e);
          } finally {
              try {
                  this.da.cleanUp(connection, ps, null);
              } catch (Exception e)  {
                  throw new DataAccessException("Error adding local items", e);
              }
          }

    }

    /**
     * delete local items
     *
     * @param sourceURI
     * @param principal
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public void deleteItems(Item[] items, String sourceURI, long principal)
    throws DataAccessException {

         Connection         connection  = null ;
         PreparedStatement  ps          = null ;

         String             id          = null ;

         try {

             connection = this.da.getConnection();

             for (int i=0, l = items.length; i < l; i++) {

                   ps = connection.prepareStatement(SQL_DELETE_LOCAL_ITEM);

                   id            = items[i].getId()           ;

                   ps.setString    (1, id         ) ;
                   ps.setString    (2, sourceURI  ) ;
                   ps.setLong      (3, principal  ) ;

                   ps.executeUpdate();

                   ps.close() ;
                   ps = null  ;

             }

         } catch (Exception e) {
             throw new DataAccessException("Error deleting local items", e);
         } finally {
             try {
                 this.da.cleanUp(connection, ps, null);
             } catch (Exception e)  {
                 throw new DataAccessException("Error deleting local items", e);
             }
         }

    }

    /**
     * get twin
     *
     * @param crc
     * @param principalId
     * @param sourceURI
     *
     * @return twin
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public Item[] getTwins(long crc         ,
                           long principalId ,
                           String sourceURI)
    throws DataAccessException {

         Connection         connection     = null ;
         PreparedStatement  ps             = null ;
         ResultSet          rs             = null ;

         ArrayList          rows           = null ;

         Item               item           ;

         int                count          = 0    ;

         try {

             connection = this.da.getConnection();

             rows = new ArrayList();

             ps = connection.prepareStatement(
                    SQL_GET_TWIN);

             ps.setString (1, sourceURI   );
             ps.setLong   (2, principalId );
             ps.setLong   (3, crc         );

             rs = ps.executeQuery();

             while(rs.next()) {
                 item = new Item(rs.getString(1),
                                 rs.getString(2),
                                 new Date(rs.getTimestamp(3).getTime()),
                                 rs.getLong(4));

                 rows.add(item);
                 count++;
             }

         } catch (Exception e) {
             throw new DataAccessException("Error getting twin local item", e);
         } finally {
             try {
                 this.da.cleanUp(connection, ps, rs);
             } catch (Exception e)  {
                 throw new DataAccessException("Error getting twin local item", e);
             }
         }

         return (Item[])rows.toArray(new Item[count]);

    }


    /**
     * get all sourceURI type items from Exchange Server
     * @param exchangeServerName
     * @param userName
     * @param credentials
     *
     * @return array of find items
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public int getExchangeAccessStatus(String exchangeServerName  ,
                                       String userName            ,
                                       String credentials         )
    throws DataAccessException {


        String webDavHeaderMsg = null ;
        String msg             = null ;

        int    status          = 0    ;

        webDavHeaderMsg
            = MessageFormat.format(ItemDAO.WEBDAV_HEADER_PROPFIND_ACCESS_STATUS,
                new Object[] {
                        "/"                 +
                        exchangeServerName  +
                        "/"                 +
                        userName            +
                        "/"
              }
        );

        try {
            msg = this.webDavHttp.sendRequest (webDavHeaderMsg,
                                               credentials,
                                               WEBDAV_MSG_ACCESS_STATUS,
                                               FILE_ENCODING);
        } catch (Exception e) {
            throw new DataAccessException("Error getting Exchange server response",
                                           e);
        }


        try {
            status = getStatusFromResponse(msg);
        } catch (Exception e) {

            throw new DataAccessException("USER "                    +
                                          userName                   +
                                          " access exchange items"   ,
                                          e)                         ;
        }

        return status;

    }

    /**
     * Build a clause
     * to webdav select request
     * between ids item array
     *
     * @param ids
     * @return webdav select clause
     **/
    public static String getClause (String[] ids) {

        String clause = null;

        int l = ids.length;

        if (l == 0) {
            return "";
        }

        clause = "where ("            +
                 CLAUSE_REPLID_FIELD  +
                 " = '"               +
                 CONTENT_REPLID_START +
                 ids[0]               +
                 "'"                  ;

        if (l > 1) {

            for (int i = 1; i < l; i++) {
                clause = clause       +
                " OR "                +
                CLAUSE_REPLID_FIELD   +
                " = '"                +
                CONTENT_REPLID_START  +
                ids[i]                +
                "'"                   ;
            }

        }

        clause = clause + ")\r\n";

        return clause;

    }

    /**
     * Make a id from replid webdavtag.
     *
     * @param replUid
     * @return id
     * @throws XmlParseException
     **/
    public static String getIdFromReplUid(String replUid)
    throws XmlParseException {

        int start = 0;

        start = replUid.indexOf(CONTENT_REPLID_START);

        if (start == -1) {
            throw new XmlParseException("Error in replUid-TAG, value: " + replUid);
        }

        return replUid.substring(start + CONTENT_REPLID_START.length());
    }

    //-------------------------------------------------------- Protected methods
    /**
     * get status from response message
     *
     * @param msg
     *
     * @throws ExchangeAccessException in case of other errors
     *
     **/
    protected int getStatusFromResponse(String msg)
    throws ExchangeAccessException
    {

        String header = null;
        String value  = null;
        int    status = 0   ;

        char[] newLine = {13, 10};

        if (msg == null && msg.length() < 30) {
            throw new ExchangeAccessException
                ("Exchange server, wrong response: " + msg);
        }

        header = msg.substring(0, msg.indexOf(new String(newLine)));

        value = header.substring(header.indexOf(" ") + 1);
        value = value.substring(0, value.indexOf(" "));

        try {
            status = Integer.parseInt(value);
        } catch  (Exception e) {
            throw new ExchangeAccessException
                ("Exchange server, wrong status: " + value);
        }

        return status;

    }

    /**
     * Check response message from Exchange server
     *
     * @param status the status to be checked
     *
     * @throws AuthenticationException in case the request has not been
     *                                 authenticated by the server
     * @throws ExchangeAccessException in case of other errors
     *
     **/
    protected void checkResponseStatus(int status)
    throws AuthenticationException ,
           ExchangeAccessException ,
           NotFoundException
    {

        if (status == STATUS_ACCESS_DANIED) {
            throw new AuthenticationException
                ("Exchange server, access denied");
        }

        if (!(status == STATUS_OK_LOAD_CHANGE ) &&
            !(status == STATUS_OK_DELETE      ) &&
            !(status == STATUS_NOT_FOUND      ) ) {
            throw new ExchangeAccessException
                ("Exchange server, error accessing resource");
        }

        if (status == STATUS_NOT_FOUND) {
            throw new NotFoundException
                ("Exchange server, resource not found");
        }

    }

    protected String getServerFromExchangeFolder(String uri) {
        return uri.substring(0, uri.indexOf("/"));
    }

    protected String getResourceFromExchangeFolder(String uri) {
        return uri.substring(uri.indexOf("/")  + 1);
    }

}
