/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.items.task.manager;

import com.funambol.framework.server.store.NotFoundException;

import com.funambol.exchange.DataAccessException;
import com.funambol.exchange.items.common.dao.ItemDAO;
import com.funambol.exchange.items.common.model.Item;
import com.funambol.exchange.items.task.dao.TaskDAO;
import com.funambol.exchange.items.task.model.Task;
import com.funambol.exchange.util.DateTools;
import com.funambol.exchange.util.StringTools;

/*
 * This class defines methods to access task data in exchange server datastore
 *
 * @version $Id: TaskManager.java,v 1.1 2006/08/08 07:51:06 gibi Exp $
 *
 **/
public class TaskManager {

    //------------------------------------------------------------- Private Data
    private TaskDAO td = null ;

    private ItemDAO id = null ;

    //--------------------------------------------------------------- Costructor

    public TaskManager(String host, int port, String exchangeServer)
    throws DataAccessException {

        this.td = new TaskDAO (host, port, exchangeServer);

        this.id = new ItemDAO (host, port, exchangeServer);
    }

    //----------------------------------------------------------- Public methods

    /**
     * get all items about principalId, sourceURI
     *
     * @param principalId
     * @param sourceURI
     *
     * @return items about principalId, sourceURI
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public Item[] getAllItems(String   sourceURI,
                              long     principalId)
    throws DataAccessException {

        return  id.getLocalItems(sourceURI, principalId);
    }

    /**
     * get tasktwin item array
     *
     * @param task the task object
     * @param principalId
     * @param sourceURI
     *
     * @return task twin item array
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public Item[] getTwinItems(Task     task        ,
                               long      principalId ,
                               String   sourceURI)
    throws DataAccessException {

        String date;
        String dueDate;

        try {

            if(task.getDate() != null) {
                date = DateTools.dateToPDIDate(task.getDate());
            } else {
                date = "";
            }

            if(task.getDueDate() != null) {
                dueDate = DateTools.dateToPDIDate(task.getDueDate());
            } else {
                dueDate = "";
            }

        } catch (Exception e) {
            throw new DataAccessException (e.getMessage());
        }

        long crc = StringTools.getCRCFromString(task.getSubject() +
                                                date              +
                                                dueDate);

        return  id.getTwins(crc,principalId, sourceURI);
    }

    /**
     * get task by id
     *
     * @param id
     * @param username
     * @param credentials
     *
     * @return find task
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public Task getTaskById(String username       ,
                            String credentials    ,
                            String id             ,
                            String exchangeFolder )
        throws DataAccessException, NotFoundException {

        Task[] tasks;

        tasks = this.td.getTasks(username          ,
                                 credentials       ,
                                 new String[] {id} ,
                                 exchangeFolder    );

        if (tasks == null || !(tasks.length > 0)) {
            throw new NotFoundException("Task with id " + id + " not found");
        }

        return tasks[0];

    }

    /**
     * get Task array by username, credentials, ids, exchangeFolder
     *
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     *
     * @return find Task array
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public Task[] getTasks(String username       ,
                           String credentials    ,
                           String ids[]          ,
                           String exchangeFolder )
        throws DataAccessException {

        Task[] tasks;

        tasks = this.td.getTasks(username        ,
                                 credentials     ,
                                 ids             ,
                                 exchangeFolder);

        return tasks;

    }

    /**
     * get Task array with setting fields only for twins
     *
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     *
     * @return array of find items
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public Task[] getTasksForTwin(String   username       ,
                                  String   credentials    ,
                                  String[] ids            ,
                                  String   exchangeFolder )
    throws DataAccessException {

        return td.getTasksForTwin(username    ,
                                  credentials ,
                                  ids         ,
                                  exchangeFolder);
    }

    /**
     * add / update task
     *
     * @param task
     * @param username
     * @param credentials
     * @param exchangeFolder
     *
     * @return new tsk
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public Task setTask(Task    task           ,
                        String  username       ,
                        String  credentials    ,
                        String  exchangeFolder )
        throws DataAccessException {

        return this.td.setTask(task           ,
                               username       ,
                               credentials    ,
                               exchangeFolder );

    }

    /**
     * remove task
     *
     * @param task
     * @param username
     * @param principal
     * @param exchangeFolder
     *
     * @throws com.funambol.exchange.DataAccessException
     **/
    public void removeTask(Task     task           ,
                           String   username       ,
                           String   principal      ,
                           String   exchangeFolder )
        throws DataAccessException {

        this.td.removeTask(task, username, principal, exchangeFolder);

    }

}
