/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.items.task.model;

import java.text.SimpleDateFormat;
import java.text.ParseException;

import java.util.Date;
import java.util.TimeZone;

import com.funambol.exchange.items.task.TaskParseException;
import com.funambol.exchange.util.DateTools;
import com.funambol.exchange.xml.XmlParser;
import com.funambol.exchange.xml.XmlParseException;

/**
 * Represent an Exchange Task
 *
 * @version $Id: Task.java,v 1.1 2006/08/08 07:51:18 gibi Exp $
 */
public class Task {

    //---------------------------------------------------------------- Constants

    // Define a common initial size for the StringBuffer.
    protected static final int INITIAL_SIZE_SBUFFER = 2048; // 2K

    private static final String TAG_PP_ACTUAL_WORK = "ActualWork";
    private static final String TAG_PP_BILLING_INFORMATION = "BillingInformation";
    private static final String TAG_PP_CATEGORIES = "Categories";
    private static final String TAG_PP_COMPANIES = "Companies";
    private static final String TAG_PP_COMPLETE = "Complete";
    private static final String TAG_PP_DATE_COMPLETED = "DateCompleted";
    private static final String TAG_PP_DUE_DATE = "DueDate";
    private static final String TAG_PP_IMPORTANCE = "Importance";
    private static final String TAG_PP_MILEAGE = "Mileage";
    private static final String TAG_PP_NOTE = "Body";
    private static final String TAG_PP_OWNER = "Owner";
    private static final String TAG_PP_PERCENT_COMPLETE = "PercentComplete";
    private static final String TAG_PP_REMINDER_SET = "ReminderSet";
    private static final String TAG_PP_REMINDER_TIME = "ReminderTime";
    private static final String TAG_PP_SENSITIVITY = "Sensitivity";
    private static final String TAG_PP_DATE = "StartDate";
    private static final String TAG_PP_STATUS = "Status";
    private static final String TAG_PP_SUBJECT = "Subject";
    private static final String TAG_PP_TEAM_TASK = "TeamTask";
    private static final String TAG_PP_TOTAL_WORK = "TotalWork";

    //------------------------------------------------------------- Private data
    private Integer actualWork = null;
    private String billingInformation = null;
    private String categories = null;
    private String companies = null;
    private boolean complete = false;
    private Date date = null;
    private Date dateCompleted = null;
    private Date dueDate = null;
    private String importance = null;
    private String mileage = null;
    private String owner = null;
    private String percentComplete = null;
    private boolean reminderSet = false;
    private Date reminderTime = null;
    private String sensitivity = null;
    private String status = null;
    private String subject = null;
    private boolean teamTask = false;
    private String textDescription = null;
    private Integer totalWork = null;

    private String href = null;

    private String username = null;
    private String id = null;

    private Date lastModified = null;
    private char state = ' ';

    //------------------------------------------------------------- Constructors

    /**
     * Creates a Task
     */
    public Task() {
    }

    /**
     * Creates a Task
     *
     * @param id client item id
     */
    public Task(String id) {
        this.id = id;
    }

    /**
     * Creates a Task
     *
     * @param id client item id
     * @param t  client item timestamp
     */
    public Task(String id, Date t) {
        this.id = id;
        this.lastModified = t;
    }

    /**
     * Creates a Task
     *
     * @param id    client item id
     * @param state client item state
     */
    public Task(String id, char state) {
        this.id = id;
        this.state = state;
    }

    /**
     * Creates a Task
     *
     * @param id
     * @param href
     * @param state
     * @param lastModified
     */
    public Task(String id,
                String href,
                char state,
                Date lastModified) {

        this.id = id;
        this.href = href;
        this.state = state;
        this.lastModified = lastModified;
    }

    /**
     * Creates a Task
     *
     * @param id      client item id
     * @param content client item content
     * @param t       client item timestamp
     */
    public Task(String id, String content, Date t) throws TaskParseException {
        getPPTask(id, content, t);
    }

    //----------------------------------------------------------- Public methods
    public String getId() {
        return this.id;
    }

    public String getUsername() {
        return this.username;
    }

    public Date getDate() {
        return this.date;
    }

    public String getHref() {
        return this.href;
    }

    public Date getLastModified() {
        return this.lastModified;
    }

    public char getState() {
        return this.state;
    }

    public Integer getActualWork() {
        return this.actualWork;
    }

    public String getBillingInformation() {
        return this.billingInformation;
    }

    public String getCompanies() {
        return this.companies;
    }

    public String getCategories() {
        return categories;
    }

    public Date getDateCompleted() {
        return this.dateCompleted;
    }

    public Date getDueDate() {
        return dueDate;
    }

    public String getImportance() {
        return importance;
    }

    public String getMileage() {
        return mileage;
    }

    public String getOwner() {
        return this.owner;
    }

    public String getPercentComplete() {
        return percentComplete;
    }

    public boolean getReminderSet() {
        return reminderSet;
    }

    public Date getReminderTime() {
        return reminderTime;
    }

    public String getSensitivity() {
        return sensitivity;
    }

    public String getStatus() {
        return status;
    }

    public String getSubject() {
        return this.subject;
    }

    public boolean getTeamTask() {
        return this.teamTask;
    }

    public boolean isTeamTask() {
        return this.teamTask;
    }

    public String getTextDescription() {
        return this.textDescription;
    }

    public Integer getTotalWork() {
        return this.totalWork;
    }

    public boolean isComplete() {
        return this.complete;
    }

    public void setId(String id) {
        this.id = id;
    }

    public void setUsername(String username) {
        this.username = username;
    }

    public void setState(char state) {
        this.state = state;
    }

    public void setDate(Date date) {
        this.date = date;
    }

    public void setHref(String href) {
        this.href = href;
    }

    public void setLastModified(Date lastModified) {
        this.lastModified = lastModified;
    }

    public void setActualWork(Integer actualWork) {
        this.actualWork = actualWork;
    }

    public void setBillingInformation(String billingInformation) {
        this.billingInformation = billingInformation;
    }

    public void setCompanies(String companies) {
        this.companies = companies;
    }

    public void setComplete(boolean complete) {
        this.complete = complete;
    }

    public void setCategories(String categories) {
        this.categories = categories;
    }

    public void setDateCompleted(Date dateCompleted) {
        this.dateCompleted = dateCompleted;
    }

    public void setDueDate(Date dueDate) {
        this.dueDate = dueDate;
    }

    public void setImportance(String importance) {
        this.importance = importance;
    }

    public void setMileage(String mileage) {
        this.mileage = mileage;
    }

    public void setOwner(String owner) {
        this.owner = owner;
    }

    public void setPercentComplete(String percentComplete) {
        this.percentComplete = percentComplete;
    }

    public void setReminderSet(boolean reminderSet) {
        this.reminderSet = reminderSet;
    }

    public void setReminderTime(Date reminderTime) {
        this.reminderTime = reminderTime;
    }

    public void setSensitivity(String sensitivity) {
        this.sensitivity = sensitivity;
    }

    public void setStatus(String status) {
        this.status = status;
    }

    public void setSubject(String subject) {
        this.subject = subject;
    }

    public void setTeamTask(boolean teamTask) {
        this.teamTask = teamTask;
    }

    public void setTextDescription(String textDescription) {
        this.textDescription = textDescription;
    }

    public void setTotalWork(Integer totalWork) {
        this.totalWork = totalWork;
    }

    /**
     * Set a Task properties from Client format content
     *
     * @param id      client item id
     * @param content client item content
     * @param t       client item timestamp
     */
    public void getPPTask(String id, String content, Date t)
            throws TaskParseException {

        String tmp;

        this.id = id;
        this.lastModified = t;

        try {

            this.date = convertDayFormatStringInDate(XmlParser.getXMLInitTagValue(content, TAG_PP_DATE));
            tmp = XmlParser.getXMLInitTagValue(content, TAG_PP_ACTUAL_WORK);
            if (tmp != null && tmp.length() > 0) {
                this.actualWork = Integer.valueOf(tmp);
            }

            this.billingInformation = XmlParser.getXMLInitTagValue(content, TAG_PP_BILLING_INFORMATION);
            this.categories = XmlParser.getXMLInitTagValue(content, TAG_PP_CATEGORIES);
            this.companies = XmlParser.getXMLInitTagValue(content, TAG_PP_COMPANIES);
            this.dateCompleted = convertDayFormatStringInDate(XmlParser.getXMLInitTagValue(content, TAG_PP_DATE_COMPLETED));
            this.dueDate = convertDayFormatStringInDate(XmlParser.getXMLInitTagValue(content, TAG_PP_DUE_DATE));
            this.importance = XmlParser.getXMLInitTagValue(content, TAG_PP_IMPORTANCE);
            this.mileage = XmlParser.getXMLInitTagValue(content, TAG_PP_MILEAGE);
            this.owner = XmlParser.getXMLInitTagValue(content, TAG_PP_OWNER);
            this.percentComplete = XmlParser.getXMLInitTagValue(content, TAG_PP_PERCENT_COMPLETE);

            this.complete = sifValueToBoolean(XmlParser.getXMLInitTagValue(content, TAG_PP_COMPLETE));
            this.reminderSet = sifValueToBoolean(XmlParser.getXMLInitTagValue(content, TAG_PP_REMINDER_SET));

            this.teamTask = sifValueToBoolean(XmlParser.getXMLInitTagValue(content, TAG_PP_TEAM_TASK));


            this.subject = XmlParser.getXMLInitTagValue(content, TAG_PP_SUBJECT);
            this.textDescription = XmlParser.getXMLInitTagValue(content, TAG_PP_NOTE);


            tmp = XmlParser.getXMLInitTagValue(content, TAG_PP_REMINDER_TIME);
            if (tmp != null && !tmp.equals("")) {
                this.reminderTime = DateTools.clientDateToDate(tmp);
            }
            this.sensitivity = XmlParser.getXMLInitTagValue(content, TAG_PP_SENSITIVITY);
            this.status = XmlParser.getXMLInitTagValue(content, TAG_PP_STATUS);

            tmp = XmlParser.getXMLInitTagValue(content, TAG_PP_TOTAL_WORK);

            if (tmp != null && tmp.length() > 0) {
                this.totalWork = Integer.valueOf(tmp);
            }

        } catch (NumberFormatException e) {
            throw new TaskParseException(e.getMessage());
        } catch (XmlParseException e) {
            throw new TaskParseException(e.getMessage());
        } catch (Exception e) {
            throw new TaskParseException(e.getMessage());
        }

    }

    /**
     * Set a Task properties to Client format content
     *
     * @return content in client format
     */
    public String toXml() throws TaskParseException {

        StringBuffer xmlMsg = new StringBuffer(INITIAL_SIZE_SBUFFER);
        String date;

        try {

            xmlMsg.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?><task>\n");

            xmlMsg.append("<")
                    .append(TAG_PP_ACTUAL_WORK)
                    .append(">")
                    .append(String.valueOf(actualWork))
                    .append("</")
                    .append(TAG_PP_ACTUAL_WORK)
                    .append(">");

            if (categories != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_CATEGORIES)
                        .append(">")
                        .append(categories)
                        .append("</")
                        .append(TAG_PP_CATEGORIES)
                        .append(">");
            }

            if (companies != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_COMPANIES)
                        .append(">")
                        .append(companies)
                        .append("</")
                        .append(TAG_PP_COMPANIES)
                        .append(">");
            }

            xmlMsg.append("<")
                    .append(TAG_PP_COMPLETE)
                    .append(">")
                    .append(booleanToSIFValue(complete))
                    .append("</")
                    .append(TAG_PP_COMPLETE)
                    .append(">");

            date = convertDateToDayFormatString(this.date);

            if (date != null && date.length() > 0) {
                xmlMsg.append("<")
                        .append(TAG_PP_DATE)
                        .append(">")
                        .append(date)
                        .append("</")
                        .append(TAG_PP_DATE)
                        .append(">");
            } else {
                xmlMsg.append("<")
                        .append(TAG_PP_DATE)
                        .append(">")
                        .append("</")
                        .append(TAG_PP_DATE)
                        .append(">");
            }

            date = convertDateToDayFormatString(dateCompleted);
            if (date!=null && date.length()>0) {
                xmlMsg.append("<")
                        .append(TAG_PP_DATE_COMPLETED)
                        .append(">")
                        .append(date)
                        .append("</")
                        .append(TAG_PP_DATE_COMPLETED)
                        .append(">");
            }

            date = convertDateToDayFormatString(dueDate);
            if (date!=null && date.length()>0) {
                xmlMsg.append("<")
                        .append(TAG_PP_DUE_DATE)
                        .append(">")
                        .append(date)
                        .append("</")
                        .append(TAG_PP_DUE_DATE)
                        .append(">");
            } else {
                xmlMsg.append("<")
                        .append(TAG_PP_DUE_DATE)
                        .append(">")
                        .append("</")
                        .append(TAG_PP_DUE_DATE)
                        .append(">");
            }

            if (importance != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_IMPORTANCE)
                        .append(">")
                        .append(importance)
                        .append("</")
                        .append(TAG_PP_IMPORTANCE)
                        .append(">");
            }

            if (mileage != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_MILEAGE)
                        .append(">")
                        .append(mileage)
                        .append("</")
                        .append(TAG_PP_MILEAGE)
                        .append(">");
            }

            if (owner != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_OWNER)
                        .append(">")
                        .append(owner)
                        .append("</")
                        .append(TAG_PP_OWNER)
                        .append(">");
            }

            if (percentComplete != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_PERCENT_COMPLETE)
                        .append(">")
                        .append(percentComplete)
                        .append("</")
                        .append(TAG_PP_PERCENT_COMPLETE)
                        .append(">");
            }

            xmlMsg.append("<")
                    .append(TAG_PP_REMINDER_SET)
                    .append(">")
                    .append(booleanToSIFValue(reminderSet))
                    .append("</")
                    .append(TAG_PP_REMINDER_SET)
                    .append(">");

            if (reminderTime != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_REMINDER_TIME)
                        .append(">")
                        .append(DateTools.toClientDate(reminderTime))
                        .append("</")
                        .append(TAG_PP_REMINDER_TIME)
                        .append(">");
            } else {
                xmlMsg.append("<")
                        .append(TAG_PP_REMINDER_TIME)
                        .append(">")
                        .append("</")
                        .append(TAG_PP_REMINDER_TIME)
                        .append(">");
            }

            if (sensitivity != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_SENSITIVITY)
                        .append(">")
                        .append(sensitivity)
                        .append("</")
                        .append(TAG_PP_SENSITIVITY)
                        .append(">");
            }

            if (status != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_STATUS)
                        .append(">")
                        .append(status)
                        .append("</")
                        .append(TAG_PP_STATUS)
                        .append(">");
            }

            if (subject != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_SUBJECT)
                        .append(">")
                        .append(subject)
                        .append("</")
                        .append(TAG_PP_SUBJECT)
                        .append(">");
            }

            xmlMsg.append("<")
                    .append(TAG_PP_TEAM_TASK)
                    .append(">")
                    .append(booleanToSIFValue(this.teamTask))
                    .append("</")
                    .append(TAG_PP_TEAM_TASK)
                    .append(">");

            if (textDescription != null) {
                xmlMsg.append("<")
                        .append(TAG_PP_NOTE)
                        .append(">")
                        .append(textDescription)
                        .append("</")
                        .append(TAG_PP_NOTE)
                        .append(">");
            }

            xmlMsg.append("<")
                    .append(TAG_PP_TOTAL_WORK)
                    .append(">")
                    .append(String.valueOf(totalWork))
                    .append("</")
                    .append(TAG_PP_TOTAL_WORK)
                    .append(">");

            xmlMsg.append("</task>");

        } catch (Exception e) {
            throw new TaskParseException("Error composing task Device: " +
                    e.getMessage());
        }

        return xmlMsg.toString();
    }


    /**
     * <p>Convert a String represetation of a date having the format yyyy-MM-dd to a Date</p>
     *
     * @param sDate the input String representation of a date
     * @return the obtained Date
     */
    private Date convertDayFormatStringInDate(String sDate) throws ParseException {
        if (sDate == null || sDate.equals("")) {
            return null;
        }
        SimpleDateFormat formatter;
        String formatDay = "yyyy-MM-dd";
        formatter = new SimpleDateFormat(formatDay);
        formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
        return formatter.parse(sDate);
    }

    private String convertDateToDayFormatString(Date date)
    {
        if (date==null)
        {
            return null;
        }

        SimpleDateFormat formatter;

        String formatDay = "yyyy-MM-dd";
        formatter = new SimpleDateFormat(formatDay);
        formatter.setTimeZone(TimeZone.getTimeZone("GMT"));

        return formatter.format(date);
    }

    /**
     * Returns the SIF representation of a boolean value.
     *
     * @param value boolean
     * @return "0" if the given boolean is false, "1" otherwise
     */
    private String booleanToSIFValue(boolean value) {
        if (value) {
            return "1";
        } else {
            return "0";
        }
    }

    /**
     * If the given string is "1" returns true, otherwise returns false
     *
     * @param value String
     * @return boolean
     */
    private boolean sifValueToBoolean(String value) {
        return value != null && "1".equals(value);
    }
}
