/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.util;

import java.util.StringTokenizer;
import java.util.zip.CRC32;

/**
 * Utility class that groups string manipulation functions.
 *
 * @version $Id: StringTools.java,v 1.1 2006/08/08 07:46:13 gibi Exp $
 */
public class StringTools {

    private static java.util.Map specCharsMap = new java.util.HashMap();
    static
    {
        specCharsMap.put("+7E", "!");
        specCharsMap.put("+60", "`");
        specCharsMap.put("+21", "!");
        specCharsMap.put("+40", "@");
        specCharsMap.put("+23", "#");
        specCharsMap.put("+24", "$");
        specCharsMap.put("+25", "%");
        specCharsMap.put("+5E", "^");
        specCharsMap.put("+26", "&");
        specCharsMap.put("+2A", "*");
        specCharsMap.put("+26", "");
        specCharsMap.put("+28", "(");
        specCharsMap.put("+29", ")");
        specCharsMap.put("+5F", "_");
        specCharsMap.put("+2B", "+");
        specCharsMap.put("+7B", "{");
        specCharsMap.put("+7D", "}");
        specCharsMap.put("+5B", "[");
        specCharsMap.put("+5D", "]");
        specCharsMap.put("+7C", "|");
        specCharsMap.put("+5C", "\\");
        specCharsMap.put("+3B", ";");
        specCharsMap.put("+22", "\"");
        specCharsMap.put("+27", "'");
        specCharsMap.put("+3C", "<");
        specCharsMap.put("+3D", ">");
        specCharsMap.put("+2C", ",");
        specCharsMap.put("+2E", ".");
        specCharsMap.put("+3F", "?");
        specCharsMap.put("_", "/");



    }

    /**
     * Splits a comma separated values string into an array of strings.
     *
     * @param s the comma separated values list - NOT NULL
     *
     * @return the elements in the list as an array
     */
    public static String[] split(String s) {
        StringTokenizer st = new StringTokenizer(s, ", ");

        String[] values = new String[st.countTokens()];

        for (int i=0; i<values.length; ++i) {
            values[i] = st.nextToken();
        }

        return values;
    }

    /**
     * Joins the given Strin[] in a comma separated String
     *
     * @param array the String[] to join - NOT NULL
     *
     * @return a comma separated list as a single string
     */
    public static String join(String[] array) {
        StringBuffer sb = new StringBuffer();
        for (int i=0; (i<array.length); ++i) {
            if (i == 0) {
                sb.append(array[i]);
            } else {
                sb.append(',').append(array[i]);
            }
        }

        return sb.toString();
    }

    /**
     * Returns true if the given string is null or zero-length, false otherwise.
     *
     * @param s the string to check
     *
     * @return true if the given string is null or zero-length, false otherwise.
     */
    public static boolean isEmpty(String s) {
        return (s == null) || (s.length() == 0);
    }

    /**
     * Replaces special characters from the given string with an underscore ('_').
     *
     * @param s the string to replace.
     *
     * @return the replaced string.
     */
    public static String replaceSpecial(String s) {
        char[] chars = s.toCharArray();
        for (int i=0; i < chars.length; ++i) {
            if ((chars[i] < '0' || chars[i]  > '9')  &&
                (chars[i]  < 'a' || chars[i]  > 'z') &&
                (chars[i]  < 'A' || chars[i]  > 'Z')) {
                chars[i] = '_';
            }
        }

        return new String(chars);
    }

    /**
     * <p>Escapes the characters in a <code>String</code> using XML entities.</p>
     *
     *
     * <p>Supports only the four basic XML entities (gt, lt, quot, amp).
     * Does not support DTDs or external entities.</p>
     *
     * @param str  the <code>String</code> to escape, may be null
     * @return a new escaped <code>String</code>, <code>null</code> if null string input
     * @see #unescapeXml(java.lang.String)
     **/
    public static String escapeXml(String str) {
        if (str != null) {
            return com.funambol.foundation.common.StringTools.escapeXml(str);
        } else {
            return null;
        }
    }

    /**
     * <p>Unescapes a string containing XML entity escapes to a string
     * containing the actual Unicode characters corresponding to the
     * escapes.</p>
     *
     * <p>Supports only the four basic XML entities (gt, lt, quot, amp).
     * Does not support DTDs or external entities.</p>
     *
     * @param str  the <code>String</code> to unescape, may be null
     * @return a new unescaped <code>String</code>, <code>null</code> if null string input
     * @see #escapeXml(String)
     **/
    public static String unescapeXml(String str) {
        if (str == null) {
            return null;
        }
        return com.funambol.foundation.common.StringTools.unescapeXml(str);
    }

    /**
     * <p>get CRC value from String</p>
     *
     *
     * @param value
     * @return CRC of value
     **/
    public static long getCRCFromString(String value) {

        CRC32 checksumEngine = new CRC32();

        value = value.toLowerCase();

        byte[] bytesMsg = value.getBytes();
        checksumEngine.update(bytesMsg, 0, bytesMsg.length);
        long checksumMessage = checksumEngine.getValue();
        checksumEngine.reset();

        return checksumMessage;
    }


    /**
     * Workaround for malformed email addressed comming from Exchange
     * @param emailAddress
     * @return the normalized email address
     */
    public static String normalizeEmailAddress(String emailAddress)
    {
        StringBuffer normalizedAddress;
        String temp;

        if (emailAddress==null)
        {
            return "";
        }

        if (!emailAddress.startsWith("IMCEA"))
        {
            return emailAddress;
        }

        emailAddress = emailAddress.substring(5, emailAddress.length());
        int i = emailAddress.indexOf('@');
        if(i>-1)
        {
            emailAddress = emailAddress.substring(0, i);
        }

        normalizedAddress = new StringBuffer();

        java.util.Set codes = specCharsMap.keySet();
        i = 0;
        while (i<emailAddress.length())
        {
            if (emailAddress.charAt(i)!='+')
            {
                normalizedAddress.append(emailAddress.charAt(i));
                i++;
                continue;
            }

            if (i+2<emailAddress.length())
            {
                temp = emailAddress.substring(i, i+3);
                if (codes.contains(temp))
                {
                    normalizedAddress.append(specCharsMap.get(temp));
                    i +=3;
                    continue;
                }

            }
            i++;
        }

        return normalizedAddress.toString();

    }

}